import json
import os
from pathlib import Path
import shutil
import subprocess
import sys
import tempfile

from . import (
    basics as poulpe,
    dataenvs as data_mod,
    errors,
)


def data_vars_key(value):
    """return the data-vars key for a value if any, None otherwise"""
    if value is not None and value.startswith("DATA-VARS:"):
        return value.split(":", 1)[1]
    return None


class UnknownBenchmarkMethod(KeyError):
    pass


def _merge_constraint(core, new):
    # XXX implement something proper sometime
    t = core.get("graph", {})
    t = t.get("visible-revision-count", {})
    core_revs = t.get("min", None)

    t = new.get("graph", {})
    t = t.get("visible-revision-count", {})
    new_revs = t.get("min", None)
    if new_revs is None:
        pass  # nothing to do
    elif core_revs is None:
        t = core.setdefault("graph", {})
        t = t.setdefault("visible-revision-count", {})
        t["min"] = new_revs
    else:
        core["graph"]["visible-revision-count"] = min(new_revs, core_revs)

    core_range = core.get("compatible-revision-range")
    new_range = new.get("compatible-revision-range")
    if new_range is not None:
        if core_range is None:
            core["compatible-revision-range"] = new_range
        else:
            rev_range = f"({core_range}) and ({new_range})"
            core["compatible-revision-range"] = rev_range


def get_benchmark(path_def: str):
    """get a benchmark from path applying possible variant selection on the way

    Expected syntax is PATH[:dimension_name=variant_name][:d=v]...
    """
    parts = str(path_def).split(":")
    path = parts[0]
    variants_definitions = parts[1:]
    variants = {}
    for d in variants_definitions:
        k, v = d.split("=")
        variants[k] = v

    data = poulpe.get_data(path)
    if data is None:
        return None
    method = data.get("meta", {}).get("method")
    if method == "simple-command":
        return SimpleCommandBenchmark(data, variants)
    elif method == "mercurial-perf-extension":
        return PerfBenchmark(data, variants)


# XXX port this over to the hg suite so we don't rely on this forever
LEGACY_RESULTS_COMPARISON = {
    "default-compare-key": "bin-env-vars.hg.changeset.node",
    "header-keys": [
        "data-env-vars.name",
        "benchmark.name",
        "bin-env-vars.hg.flavor",
        "bin-env-vars.hg.py-re2-module",
        "bin-env-vars.hg.changeset.node",
    ],
    "sort-keys": {
        "bin-env-vars.hg.changeset.node": [
            "bin-env-vars.hg.changeset.rank",
            "bin-env-vars.hg.changeset.node",
        ]
    },
}


class Benchmark:
    def __init__(self, data, variants=None):
        self._raw_data = data

        assert self._raw_data["meta"]["format"] == "0"

        self._data = self._raw_data.copy()
        if variants is None:
            variants = {}
        self._selected_variants = variants

        # find all variant and pick default one when needed.
        for dimension, variants in self.all_dimensions.items():
            selected = self._selected_variants.get(dimension)
            if selected is None:
                for name, values in variants.items():
                    if values.get("default", False):
                        selected = name
                        break
                else:
                    msg = "missing default value for dimensions: %s"
                    assert False, msg % dimension
                self._selected_variants[dimension] = selected

    def new_with_variants(self, variants):
        """create a new version of this benchmark with the selected variants"""
        return self.__class__(self._raw_data, variants=variants)

    @property
    def name(self):
        return self._data["meta"]["name"]

    def get_var(self, key, default=None):
        return poulpe.get_one_value(self._data, key, default=default)

    @property
    def selected_variants(self):
        selected = {}
        all = self.all_dimensions
        for dim, var in self._selected_variants.items():
            if dim in all:
                selected[dim] = var
        return selected

    @property
    def all_dimensions(self):
        return {}

    @property
    def _base_constraints(self):
        return {}

    @property
    def constraints(self):
        c = self._base_constraints.copy()
        c["needed-vars"] = self.needed_vars
        # XXX need to fetch graph constraint for the core variable too
        min_rev = None

        for dimension, variants in self.all_dimensions.items():
            selected = self._selected_variants[dimension]
            values = variants[selected]
            _merge_constraint(c, values.get("constraints", {}))
        return c

    @property
    def environment(self):
        env = self._data.get("environment", {})
        for dimension, variants in sorted(self.all_dimensions.items()):
            selected = self._selected_variants[dimension]
            values = variants[selected]
            var_env = values.get("environment")
            if var_env is not None:
                env.update(var_env)
        return env

    @staticmethod
    def apply_data_on_env(data_env, env):
        new_env = env.copy()
        for env_key, value in env.items():
            data_key = data_vars_key(value)
            if data_key is not None:
                new_env[env_key] = data_env.get_benchmark_input_vars(data_key)
        return new_env

    @property
    def needed_vars(self):
        nvars = set()
        for k, v in self.environment.items():
            key = data_vars_key(v)
            if key is not None:
                nvars.add(key)
        return nvars

    def compatible_with(self, bin_env, data_env) -> bool:
        """return True if this benchmark can run on these input

        This compare the constraints of this benchmark with the property and
        available variable in the binary-env and the data-env to determine if
        this instance of the benchmark is compatible with it.

        Return True if compatible False otherwise.
        """
        constraints = self.constraints
        for var in constraints.get("needed-vars", ()):
            if not data_env.has_benchmark_input_vars(var):
                return False

        graph_constraints = constraints.get("graph")
        if graph_constraints:
            rev_count = graph_constraints.get("visible-revision-count")
            if rev_count is not None:
                # XXX also handle non mercurial repository
                # XXX also handle other keys
                visible_revs = (
                    data_env.data_env_vars.get(
                        "mercurial",
                        {},
                    )
                    .get(
                        "repo",
                        {},
                    )
                    .get(
                        "graph",
                        {},
                    )
                    .get("visible-revision-count")
                )
                if visible_revs is not None:
                    if rev_count > visible_revs:
                        return False
        return True


class SimpleCommandBenchmark(Benchmark):
    def __init__(self, data, variants=None):
        super().__init__(data, variants)

        assert self._raw_data["meta"]["method"] == "simple-command"

        for dimension, variants in self.all_dimensions.items():
            selected = self._selected_variants[dimension]

            # gather the changes to the benchmark
            new_cwd = variants[selected].get("cwd")
            if new_cwd is not None:
                self._data["simple-command"]["cwd"] = new_cwd
            extend_command = variants[selected].get("extend-command")
            if extend_command is not None:
                self._data["simple-command"]["command"] += extend_command
            new_return_codes = variants[selected].get("acceptable-return-codes")
            if new_return_codes is not None:
                self._data["simple-command"]["acceptable-return-codes"] = (
                    new_return_codes
                )

    @property
    def all_dimensions(self):
        l_1 = self._data.get("simple-command", {})
        l_2 = l_1.get("variants", {})
        all_dimensions = l_2.get("dimensions", {})
        return all_dimensions.copy()

    @property
    def needed_vars(self):
        variables = super().needed_vars
        cwd = self.get_var("simple-command.cwd")
        key = data_vars_key(cwd)
        if key is not None:
            variables.add(key)
        # XXX missing the "variable" part
        return variables

    def run_one(self, bin_env_path, data_env_path, debug=False):
        data_env = data_mod.get_data_env(data_env_path)

        cmd = SimpleCommand(self)
        cmd.apply_data_env(data_env)

        prepare = self.get_var("simple-command.prepare-run")
        cleanup = self.get_var("simple-command.cleanup-run")
        gather_info = self.get_var("simple-command.gather-info")
        temp_dir = tempfile.mkdtemp()  # TODO same tmp dir as `copy-data-env`
        try:
            r, runtime_info = self._time_command(
                bin_env_path,
                data_env_path,
                cmd,
                temp_dir=temp_dir,
                prepare=prepare,
                cleanup=cleanup,
                gather_info=gather_info,
                debug=debug,
            )
        finally:
            shutil.rmtree(temp_dir, ignore_errors=True)
        res = {}

        # TODO record poulpe version in results in case results accidentally
        # get bogus for some poulpe version.

        # we should store more
        res["time"] = {}
        res["time"]["user"] = r["results"][0]["user"]
        res["time"]["system"] = r["results"][0]["system"]
        res["time"]["median"] = r["results"][0]["median"]
        res["time"]["mean"] = r["results"][0]["mean"]
        res["time"]["standard-deviation"] = r["results"][0]["stddev"]
        res["time"]["min"] = r["results"][0]["min"]
        res["time"]["max"] = r["results"][0]["max"]
        if runtime_info is not None:
            # XXX is there a better place to store this?
            res["runtime_info"] = runtime_info
        return res

    def _time_command(
        self,
        bin_env_path,
        data_env_path,
        cmd,
        temp_dir,
        prepare=None,
        cleanup=None,
        gather_info=None,
        debug=False,
    ):
        bin_env_path = os.path.abspath(bin_env_path)
        data_env_path = os.path.abspath(data_env_path)
        shell_path = poulpe.bin_env_script(bin_env_path)

        data_env = data_mod.get_data_env(data_env_path)
        our_env = self.apply_data_on_env(data_env, self.environment)
        env = os.environ.copy()
        env.update(our_env)
        # Give the benchmarks the tools to do more advanced path manipulation
        # without assuming the exact structure.
        env["POULPE_DATA_ENV_ROOT"] = data_env_path
        env["POULPE_TMP_DIR"] = temp_dir
        if cmd.cwd is not None:
            env["POULPE_CWD"] = cmd.cwd

        with tempfile.NamedTemporaryFile("w") as tmp_result:
            time_cmd = [
                shell_path,
                "hyperfine",
                "--shell",
                "/bin/bash",
                "--export-json",
                tmp_result.name,
                "--show-output",
            ]
            if cmd.acceptable_return_codes != {0}:
                # If there are non-zero status codes, we will check them in
                # post-processing.
                time_cmd.append("--ignore-failure")

            if debug:
                time_cmd.append("--runs")
                time_cmd.append("2")
                if prepare is not None:
                    prepare = (
                        f'echo "### starting prepare ###" ; date ; {prepare}'
                    )
            elif os.environ.get("RUNTESTDIR"):
                # speed up the test run
                time_cmd.append("--runs")
                # would use 1 but tht make "standard-deviation" entry disappear.
                time_cmd.append("2")
            else:
                time_cmd.append("--min-runs")
                time_cmd.append("3")

            if prepare is not None:
                if not bool(self.get_var("simple-command.no-set-builtin")):
                    prepare = f"set -eEuo pipefail; {prepare}"
                time_cmd.append("--prepare")
                time_cmd.append(prepare)

            if cleanup is not None:
                if not bool(self.get_var("simple-command.no-set-builtin")):
                    cleanup = f"set -eEuo pipefail; {cleanup}"
                # beware that `--cleanup` is global to a hyperfine invocation
                time_cmd.append("--conclude")
                time_cmd.append(cleanup)

            command = cmd.command
            if debug:
                command = f'echo "### starting command ###" ; date ; {command}'

            time_cmd.append("--")
            time_cmd.append(command)

            cwd = data_env_path
            if cmd.cwd is not None:
                cwd = os.path.join(cwd, cmd.cwd)

            if debug:
                stdout = sys.stdout
                stderr = sys.stderr
            else:
                stdout = subprocess.PIPE
                stderr = subprocess.PIPE
            runtime_info = None
            if gather_info is not None:
                if not bool(self.get_var("simple-command.no-set-builtin")):
                    gather_info = f"set -eEuo pipefail; {gather_info}"
                r = subprocess.run(
                    [shell_path, "/bin/bash", "-c", gather_info],
                    cwd=cwd,
                    env=env,
                    stdout=stdout,
                    stderr=stderr,
                )
                if r.returncode != 0:
                    raise errors.BenchmarkRunFailure(
                        command=gather_info,
                        cwd=cwd,
                        return_code=r.returncode,
                        stdout=r.stdout,
                        stderr=r.stderr,
                        message="Gathering information prior to benchmark failed",
                    )
                runtime_info_path = Path(temp_dir) / "gathered_info.json"
                try:
                    with runtime_info_path.open(mode="r") as f:
                        runtime_info = json.load(f)
                    # clean up incase it somehow affects the rest of it
                    os.unlink(runtime_info_path)
                except FileNotFoundError:
                    pass

            r = subprocess.run(
                time_cmd, cwd=cwd, env=env, stdout=stdout, stderr=stderr
            )
            if r.returncode != 0:
                raise errors.BenchmarkRunFailure(
                    command=cmd,
                    cwd=cwd,
                    return_code=r.returncode,
                    stdout=r.stdout,
                    stderr=r.stderr,
                )

            with open(tmp_result.name) as f:
                results_as_json = json.load(f)

            exit_codes = results_as_json["results"][0]["exit_codes"]
            for index, code in enumerate(exit_codes):
                if code not in cmd.acceptable_return_codes:
                    acceptable_return_codes = sorted(
                        cmd.acceptable_return_codes
                    )
                    msg = (
                        f"Run {index + 1} out of {len(exit_codes)}"
                        f" returned with exit code '{code}'.\n"
                        f"Subsequent runs may also have failed, "
                        f"expected any of '{acceptable_return_codes}'."
                    )
                    raise errors.BenchmarkRunFailure(
                        command=cmd,
                        cwd=cmd.cwd,
                        return_code=r.returncode,
                        stdout=r.stdout,
                        stderr=r.stderr,
                        message=msg,
                    )
            return results_as_json, runtime_info


class SimpleCommand(object):
    # XXX should probably be merged back (or rebalanced) with the
    # SimpleCommandBenchmark class

    def __init__(self, benchmark_data: Benchmark):
        self._benchmark_data = benchmark_data

        def get_var(key, default=None):
            key = f"simple-command.{key}"
            return benchmark_data.get_var(key, default=default)

        self.base_command = get_var("command")
        assert self.base_command is not None
        self.command = None

        self.acceptable_return_codes = set(
            get_var("acceptable-return-codes", [0])
        )

        self.cwd = get_var("cwd", ".")

    def apply_data_env(self, data_env):
        variables_data = self._benchmark_data.get_var("simple-command").get(
            "variables"
        )

        get_var = data_env.get_benchmark_input_vars

        variables = {}
        if variables_data is not None:
            for name, value in variables_data.items():
                key = data_vars_key(value)
                if key is not None:
                    value = get_var(key)
                    assert value is not None
                variables[name] = value

        self.command = self.base_command.format(**variables)

        key = data_vars_key(self.cwd)
        if key is not None:
            self.cwd = get_var(key)

    def __repr__(self):
        cmd = repr(self.command)
        return f"SimpleCommand(command={cmd}; acceptable-return-codes={self.acceptable_return_codes})"


class PerfBenchmark(Benchmark):
    """A benchmark that use the Mercurial performance extension

    It assumes an up to date mercurial repository live in

    The path to the perf extension should be contained in the HG_PERF_PATH
    environment variable.

    bin-env-dir/../repos/mercurial/

    (This might live in a pluging in the future.)
    """

    def __init__(self, data, variants=None):
        super().__init__(data, variants)

        assert self._raw_data["meta"]["method"] == "mercurial-perf-extension"

        self._data["hg-perf-ext"].setdefault("args", [])

        for dimension, variants in self.all_dimensions.items():
            selected = self._selected_variants[dimension]

            # gather the changes to the benchmark
            new_cwd = variants[selected].get("cwd")
            if new_cwd is not None:
                self._data["hg-perf-ext"]["cwd"] = new_cwd
            args = variants[selected].get("args")
            if args:
                self._data["hg-perf-ext"]["args"].extend(args)
            setup_script = variants[selected].get("setup-script")
            if setup_script:
                self._data["hg-perf-ext"]["setup-script"] = setup_script

    @property
    def all_dimensions(self):
        l_1 = self._data.get("hg-perf-ext", {})
        l_2 = l_1.get("variants", {})
        all_dimensions = l_2.get("dimensions", {})
        return all_dimensions.copy()

    @property
    def _base_constraints(self):
        return self.get_var("hg-perf-ext.constraints", {})

    @property
    def needed_vars(self):
        variables = super().needed_vars
        cwd = self.get_var("hg-perf-ext.cwd")
        key = data_vars_key(cwd)
        if key is not None:
            variables.add(key)
        args = self.get_var("hg-perf-ext.args")
        if args is not None:
            for value in args:
                key = data_vars_key(value)
                if key is not None:
                    variables.add(key)
        return variables

    def run_one(self, bin_env_path, data_env_path, debug=False):
        bin_env_path = os.path.abspath(bin_env_path)
        data_env_path = os.path.abspath(data_env_path)
        shell_path = poulpe.bin_env_script(bin_env_path)

        perf_command = self.get_var("hg-perf-ext.command")

        if "HG_PERF_PATH" not in os.environ:
            msg = "missing environment variable HG_PERF_PATH"
            raise RuntimeError(msg)
        perf_ext_path = os.environ["HG_PERF_PATH"]
        if not os.path.exists(perf_ext_path):
            msg = "could not find perf extension at: %s"
            raise RuntimeError(msg % perf_ext_path)

        data_env = data_mod.get_data_env(data_env_path)

        data_var = data_env.get_benchmark_input_vars

        cwd = data_env_path
        cfg_cwd = self.get_var("hg-perf-ext.cwd")
        if cfg_cwd is not None:
            key = data_vars_key(cfg_cwd)
            if key is not None:
                cfg_cwd = data_var(key)
            cwd = os.path.join(cwd, cfg_cwd)

        data_env = data_mod.get_data_env(data_env_path)
        our_env = self.apply_data_on_env(data_env, self.environment)
        env = os.environ.copy()
        env.update(our_env)

        setup_script = self.get_var("hg-perf-ext.setup-script")
        use_set_builtin = not bool(self.get_var("hg-perf-ext.no-set-builtin"))
        if setup_script is not None:
            if use_set_builtin:
                setup_script = f"set -eEuo pipefail; {setup_script}"
            subprocess.check_call(
                [shell_path, "bash", "-c", setup_script],
                cwd=cwd,
                env=env,
            )

        cmd = [
            shell_path,
            "hg",
            "--config",
            f"extensions.perf={perf_ext_path}",
            "--config",
            "perf.all-timing=yes",
            perf_command,
            "--template",
            "json",
        ]

        def filter_arg(value):
            key = data_vars_key(value)
            if key is not None:
                value = data_var(key)
                assert value is not None
            return value

        args = self.get_var("hg-perf-ext.args")
        args = [filter_arg(a) for a in args]

        if args:
            cmd.extend(args)

        if debug:
            stdout = sys.stdout
            stderr = sys.stderr
        else:
            stdout = subprocess.PIPE
            stderr = subprocess.PIPE
        r = subprocess.run(
            cmd,
            cwd=cwd,
            env=env,
            stdout=stdout,
            stderr=stderr,
        )
        if r.returncode != 0:
            raise errors.BenchmarkRunFailure(
                command=cmd,
                cwd=cwd,
                return_code=r.returncode,
                stdout=r.stdout,
                stderr=r.stderr,
            )

        data = json.loads(r.stderr)

        res = {}
        # XXX wall/sys/user/comb should probably be a sub-layer in 'time'

        # XXX note that the perf extension does not provides any "min"
        res["time"] = {}
        res["time"]["min"] = data[0]["wall"]
        res["time"]["mean"] = data[0]["avg.wall"]
        res["time"]["median"] = data[0]["median.wall"]
        res["time"]["max"] = data[0]["max.wall"]
        res["sys-time"] = {}
        res["sys-time"]["min"] = data[0]["sys"]
        res["sys-time"]["mean"] = data[0]["avg.sys"]
        res["sys-time"]["median"] = data[0]["median.sys"]
        res["sys-time"]["max"] = data[0]["max.sys"]
        res["user-time"] = {}
        res["user-time"]["min"] = data[0]["user"]
        res["user-time"]["mean"] = data[0]["avg.user"]
        res["user-time"]["median"] = data[0]["median.user"]
        res["user-time"]["max"] = data[0]["max.user"]
        res["comb-time"] = {}
        res["user-time"]["comb"] = data[0]["comb"]
        res["comb-time"]["mean"] = data[0]["avg.comb"]
        res["comb-time"]["median"] = data[0]["median.comb"]
        res["comb-time"]["max"] = data[0]["max.comb"]
        res["run-count"] = data[0]["count"]
        return res
