import importlib_resources as resources
import os
import subprocess
import sys

from pathlib import Path

import click

from . import compare
from . import den as den_mod
from . import filter
from . import result
from . import scheduling

BINARIES = resources.files("poulpe.bin")


# XXX hack to delay the integration of the various script
def find_sub_command():
    cmds = []
    for f in BINARIES.iterdir():
        if (
            f.is_file()
            and f.is_file
            and os.access(f, os.X_OK)
            and not f.name.startswith(".")
        ):
            cmds.append(f.name)
    return cmds


def filter_option(target, option_prefix=""):
    option = "--filter"
    argument = "filters"
    if option_prefix:
        option = f"--{option_prefix}-filter"
        arg_prefix = option_prefix.replace("-", "_")
        argument = f"{arg_prefix}_filters"

    return click.option(
        option,
        argument,
        multiple=True,
        help=f"""Criteria to filter {target} with. Filters are additive.

They are of the form `key=value`, where `key` is the dot-delimited key
inside the result file, and `value` is the expected value.
If `value` starts with `re:`, then it is checked against
the regular expression that follows.""",
    )


@click.group()
def poulpe_cli():
    """The main entry point for Poulpe command"""
    pass


def benchmark_name_filter_autocomplete(ctx: click.Context, param, incomplete):
    den_root = Path(ctx.params["den_path"])
    den = den_mod.PoulpeDen(den_root)
    return [
        b.name
        for (p, b) in den.find_benchmarks([incomplete] if incomplete else None)
    ]


all_sub_commands = []


# dynamically declare the scripts
for cmd in find_sub_command():

    @poulpe_cli.command(
        name=cmd,
        context_settings={
            "ignore_unknown_options": True,
            "allow_interspersed_args": False,
        },
    )
    @click.option("--help", is_flag=True)
    @click.argument("subcommand_args", nargs=-1, type=click.UNPROCESSED)
    def subcommand(help, subcommand_args, subcommand=cmd):
        subcommand_file = BINARIES.joinpath(subcommand)

        if not subcommand_file.is_file():
            m = f"Subcommand `{subcommand}` does not exist."
            raise click.BadParameter(m)

        if help:
            subcommand_args = ["--help"]

        with resources.as_file(subcommand_file) as f:
            full_cmd = [str(f)] + list(subcommand_args)
            exit(subprocess.call(full_cmd))

    subcommand.help = f"use `poulpe {cmd} --help` for details"

    all_sub_commands.append(subcommand)


@poulpe_cli.command()
@click.option("--den-path", type=click.Path(exists=True), default=".")
def den_check(den_path="."):
    """Verify and Summarise the state of a Poulpe Den"""
    den_root = Path(den_path)
    den = den_mod.PoulpeDen(den_root)
    den_mod.check(den)


@poulpe_cli.command()
@click.option("--list-variants", is_flag=True)
@click.option("--show-path", is_flag=True)
@click.option("--den-path", type=click.Path(exists=True), default=".")
@click.argument(
    "pattern", nargs=-1, shell_complete=benchmark_name_filter_autocomplete
)
def den_list_benchmarks(
    pattern, den_path=".", list_variants=False, show_path=False
):
    """Find and list benchmarks available in a Poulpe Den

    Optional patterns can be passed to filter on benchmark name.
    Pattern starting with "re:" will be matched as a regular expression.
    """
    den_root = Path(den_path)
    den = den_mod.PoulpeDen(den_root)
    den_mod.list_benchmarks(
        den,
        list_variants=list_variants,
        show_path=show_path,
        pattern=pattern if pattern else None,
    )


@poulpe_cli.command()
@click.option("--den-path", type=click.Path(exists=True), default=".")
@click.option("--list-vars", is_flag=True)
@filter_option(target="data-environments")
def den_list_data_envs(den_path=".", list_vars=False, filters=None):
    """Find and list Data Enviroment available in a Poulpe Den"""
    den_root = Path(den_path)
    den = den_mod.PoulpeDen(den_root)
    if filters is None:
        filters = []
    filters = filter.parse_filters(filters)
    den_mod.list_data_env(den, list_vars=list_vars, filters=filters)


@poulpe_cli.command()
@click.option("--den-path", type=click.Path(exists=True), default=".")
def den_list_bin_env_setups(den_path="."):
    """Find and list automatic Binary Environment automated setup in a Poulpe Den"""
    den_root = Path(den_path)
    den = den_mod.PoulpeDen(den_root)
    den_mod.list_bin_env_setup(den)


@poulpe_cli.command()
@click.option("--den-path", type=click.Path(exists=True), default=".")
def den_interactive_bin_env_spec(den_path="."):
    """Interactive command to guide you though creating a valid bin-env spec
    from the available suites"""
    den_root = Path(den_path)
    den = den_mod.PoulpeDen(den_root)
    den_mod.interactive_bin_env_spec(den)


@poulpe_cli.command(no_args_is_help=True)
@click.option("--den-path", type=click.Path(exists=True), default=".")
@click.argument(
    "bin-env-spec",
)
def den_setup_bin_env(
    bin_env_spec,
    den_path=".",
):
    """Setup a binary environment from the spec, reusing one if it exists.

    Use `poulpe den-interactive-bin-env-spec` to get help creating a
    bin-env spec"""
    den_root = Path(den_path)
    den = den_mod.PoulpeDen(den_root)
    bin_env_spec = scheduling.BinEnvSpec.from_piece(bin_env_spec)
    bin_env = scheduling.ensure_bin_env(den, bin_env_spec)
    if bin_env is None:
        raise click.ClickException("failed to setup bin-env")
    click.echo(bin_env)


@poulpe_cli.command()
@click.option("--den-path", type=click.Path(exists=True), default=".")
@click.option(
    "--schedule-path",
    type=click.Path(exists=True),
    default="./current.poulpe-schedule",
    help="path to a file containing task to run.",
)
@click.option(
    "--report-path",
    type=click.Path(),
    default="./current.poulpe-report",
    help="path to a file containing task being run (file is appened to).",
)
@click.option("--lazy", is_flag=True)
def den_execute_tasks(
    den_path=".",
    schedule_path="./current.poulpe-schedule",
    report_path="./current.poulpe-report",
    lazy=False,
):
    """execute the content of a schedule file"""
    den_root = Path(den_path)
    schedule_path = Path(schedule_path)
    report_path = Path(report_path)
    den = den_mod.PoulpeDen(den_root)
    schedule = scheduling.get_run_tasks(den, schedule_path)
    res = scheduling.run_multiple_tasks(den, schedule, report_path, lazy=lazy)
    (successes, setup_issues, run_issues, unknown_failures) = res

    click.secho(
        f"{successes} out of {len(schedule)} tasks have succeeded", fg="green"
    )
    if setup_issues:
        click.secho(
            f"{setup_issues} out of {len(schedule)} tasks have not run "
            "due to missing data requirements",
            err=True,
            fg="yellow",
        )
    if run_issues:
        click.secho(
            f"{run_issues} out of {len(schedule)} tasks have failed "
            "during the run",
            err=True,
            fg="red",
        )
    if unknown_failures:
        click.secho(
            f"{unknown_failures} out of {len(schedule)} tasks have failed "
            "in an unexpected way",
            err=True,
            fg="red",
        )


@poulpe_cli.command(no_args_is_help=True)
@click.option("--den-path", type=click.Path(exists=True), default=".")
@click.option("--all-variants", is_flag=True)
@filter_option(target="data-environments", option_prefix="data-env")
@click.option(
    "--benchmark-name-filter",
    "benchmark_name_filters",
    shell_complete=benchmark_name_filter_autocomplete,
    multiple=True,
)
@click.option(
    "--benchmark-variant-filter",
    "benchmark_variant_filters",
    multiple=True,
)
@click.argument("bin-env-spec", nargs=-1)
def den_generate_tasks(
    bin_env_spec,
    den_path=".",
    all_variants=False,
    data_env_filters=None,
    benchmark_name_filters=None,
    benchmark_variant_filters=None,
):
    """Generate a tasks list from parameters.


    See available bin-env options with `poulpe den-list-bin-env-setups`.

    \b
        Example:
        $ poulpe den-generate-tasks \\
            hg:MERCURIAL_VERSION=abcdef123456:MERCURIAL_FLAVOR=rust \\
            hg:MERCURIAL_VERSION=abcdef123456:MERCURIAL_FLAVOR=default


    See available benchmark options with `poulpe den-list-benchmarks`

    See available data-environment options with `poulpe den-list-data-envs`
    """
    data_env_filters = filter.parse_filters(data_env_filters)
    benchmark_name_pattern = filter.build_pattern(benchmark_name_filters)
    benchmark_variant_filters = filter.parse_filters(benchmark_variant_filters)
    den_root = Path(den_path)
    den = den_mod.PoulpeDen(den_root)
    tasks = scheduling.generate_tasks(
        den,
        bin_env_spec,
        all_variants=all_variants,
        data_env_filters=data_env_filters,
        benchmark_name_pattern=benchmark_name_pattern,
        benchmark_variant_filters=benchmark_variant_filters,
    )
    for t in tasks:
        click.echo(" ".join(t))


@poulpe_cli.command()
@click.option("--path-file", type=click.File())
@click.option("--json-file", type=click.File())
@click.option(
    "-d",
    "--dimension",
    default=None,
    help="ignore the following dimension when grouping (and displaying)",
)
@click.option(
    "-i",
    "--ignore-dimension",
    "ignored_dimensions",
    default=[],
    multiple=True,
    help="The path to the dimension along which to compare (e.g. the nodeid)",
)
@click.option(
    "-r",
    "--reference-key",
    # XXX leak of suite specific logic
    default=None,
    help="use this value as comparison baseline (when present)",
)
@click.option(
    # aggregate
    "--value-type",
    default="median",
    type=click.Choice(["median", "mean", "min", "max"]),
    help="which benchmark value type should be used",
)
@click.option(
    "--metric-type",
    default="wall",
    type=click.Choice(["wall", "user", "system"]),
    help="What metric should we display",
)
@click.option(
    "--minimal-relative-change",
    default=None,
    type=float,
    help="only display group with relative change above threshold",
)
@click.option(
    "--minimal-real-change",
    default=None,
    type=float,
    help="only display group with relative absolute above threshold",
)
def result_compare(
    path_file,
    json_file,
    dimension,
    ignored_dimensions=(),
    reference_key=None,
    value_type="median",
    metric_type="wall",
    minimal_relative_change=None,
    minimal_real_change=None,
):
    """Compare results together.

    `poulpe result-search | poulpe result-compare`"""
    if path_file is None and json_file is None and sys.stdin.isatty():
        click.echo("reading file paths from stdin", file=sys.stderr)

    compare.compare_results(
        path_file,
        json_file,
        dimension=dimension,
        ignored_dimensions=ignored_dimensions,
        reference_key=reference_key,
        value_type=value_type,
        metric_type=metric_type,
        minimal_relative_change=minimal_relative_change,
        minimal_real_change=minimal_real_change,
    )


@poulpe_cli.command()
@click.option(
    "--content-as-json",
    is_flag=True,
    default=False,
    help="Display the contents as JSON instead of the filenames",
)
@filter_option(target="results")
@click.argument("result_paths", metavar="RESULTS", nargs=-1)
def result_search(
    result_paths,
    content_as_json=False,
    filters=None,
):
    """Find result files matching filters.

    \b
    ARGS:
        <RESULTS> are paths to result files or directories containing `.pbr`
    """
    if not result_paths:
        result_paths = ["."]

    results = result.find_results(result_paths)

    if filters is None:
        filters = []
    filters = filter.parse_filters(filters)
    results = result.filter_results(filters, results)

    if content_as_json:
        result.content_as_json(results)
    else:
        for path in results:
            click.echo(path)


if __name__ == "__main__":
    poulpe_cli
