#!/usr/bin/env python3

import collections
import json
import sys

import click
from poulpe.benchmarks import LEGACY_RESULTS_COMPARISON
from poulpe.utils import join_escaped, split_unescape

from . import get_data


def get_value(key, result):
    treck = list(key[::-1])
    current = result
    while treck and current is not None:
        k = treck.pop()
        current = current.get(k)
    return current


def make_key_result(header_keys, compare_key, ignored_dimensions=()):
    headers_keys = [k for k in header_keys if k != compare_key]

    def key_result(result):
        headers = []
        for key in headers_keys:
            value = get_value(key, result)
            if value is not None:
                key = join_escaped(key, ".")
                if key in ignored_dimensions:
                    continue
                headers.append((key, value))

        variant_cmp_key = join_escaped(compare_key, ".")
        variants = result["benchmark"].get("variants", {})
        for key, value in sorted(variants.items()):
            key = f"benchmark.variants.{key}"
            if key == variant_cmp_key:
                continue
            if key in ignored_dimensions:
                continue
            headers.append((key, value))
        return tuple(headers)

    return key_result


def dispatch_results(results, dimension, ignored_dimensions=()):
    dispatched = collections.defaultdict(list)
    if dimension is not None:
        compare_key = tuple(split_unescape(dimension, "."))

    sort_keys = {}

    for r in results:
        # TODO figure out where to put this information, or how to merge it
        # from multiple sources, and which ones etc.
        # For now the bin-env-vars will do
        if "results-comparison" not in r["bin-env-vars"]:
            r["bin-env-vars"]["results-comparison"] = LEGACY_RESULTS_COMPARISON
        comparison = r["bin-env-vars"]["results-comparison"]
        if dimension is None:
            # No dimension given, we must ask each result along which dimension
            # they should be compared
            compare_key = comparison["default-compare-key"]
            compare_key = tuple(split_unescape(compare_key, "."))
        sort_keys.update(
            {
                k: [tuple(split_unescape(v1, ".")) for v1 in v]
                for k, v in comparison["sort-keys"].items()
            }
        )
        header_keys = [
            tuple(split_unescape(h, ".")) for h in comparison["header-keys"]
        ]
        key_func = make_key_result(header_keys, compare_key, ignored_dimensions)
        headers = key_func(r)
        dispatched[(headers, compare_key)].append(r)

    for (_h, comp_key), some_results in dispatched.items():
        sort_by = join_escaped(comp_key, ".")
        sort_key = sort_keys.get(sort_by)
        if sort_key is not None:

            def key(res):
                values = []
                for k in sort_key:
                    value = get_value(k, res)
                    if value is not None:
                        values.append(value)
                return values

        else:

            def key(res):
                val = get_value(compare_key, res)
                if val is None:
                    val = ""
                return val

        some_results.sort(key=key)

    return dispatched


def display_all_results(
    results,
    reference_key=None,
    value_type="median",
    metric_type="wall",
    # XXX doing the change filtering at the display level is poor architecture.
    # Right now we have these steps:
    # - group results
    # - display (with filtering)
    # We should have these steps:
    # - group result
    # - compute changes within groups
    # - filter groups
    # - display group
    minimal_relative_change=None,
    minimal_real_change=None,
):
    used_compare_key = False
    for (headers, compare_key), some_results in sorted(results.items()):
        if _display_one_result_group(
            headers,
            some_results,
            compare_key,
            reference_key=reference_key,
            value_type=value_type,
            metric_type=metric_type,
            minimal_relative_change=minimal_relative_change,
            minimal_real_change=minimal_real_change,
        ):
            used_compare_key = True
    return not used_compare_key


def _display_one_result_group(
    headers,
    results,
    compare_key,
    reference_key=None,
    value_type="median",
    metric_type="wall",
    minimal_relative_change=None,
    minimal_real_change=None,
):
    lines = []
    echo = lines.append

    if metric_type != "wall":
        # they are currently flat we should fix this at collection time
        value_type = metric_type

    key_size = max(len(k) for k, v in headers)
    prefix = click.style("###", fg="cyan")
    for k, v in headers:
        k = k.ljust(key_size)
        equal = click.style("=", fg="cyan")
        k = click.style(k, fg="magenta")
        echo(f"{prefix} {k} {equal} {v}")
        prefix = click.style("  #", fg="cyan")

    base = None
    ref_line = None
    if reference_key is not None:
        for r in results:
            key = get_value(compare_key, r)
            if compare_key[-1] == "node":
                # XXX leak of suite-specific logic
                key = key[:12]
            times = r["result"]["time"]
            if key == reference_key and value_type in times:
                base = times[value_type]
                ref_line = r

    max_abs_change = 0
    max_abs_diff = 0
    used_compare_key = False
    # display the result now
    for r in results:
        key = get_value(compare_key, r)
        if key is not None:
            used_compare_key = True
        if compare_key[-1] == "node":
            # XXX leak of suite-specific logic
            key = key[:12]

        times = r["result"]["time"]
        value = times[value_type]
        if base is None and value is not None:
            base = value
            ref_line = r

        if value is None:
            value_d = click.style(f"no value for <{value_type}>", fg="red")
        else:
            value_d = f"{value:6.6f}"

        dev = ""
        if value_type == "mean" and "standard-deviation" in times:
            std_dev = times["standard-deviation"]
            dev = f" (σ {std_dev:6.6f})"

        display = ""
        if r is ref_line:
            display = click.style(" ~~~~~", fg="blue")
        elif value is not None:
            change = value / base
            diff = value - base
            max_abs_change = max(abs(change - 1), max_abs_change)
            max_abs_diff = max(abs(diff), max_abs_change)
            if change <= 0.99:
                if change > 0.95:
                    color = "cyan"
                else:
                    color = "green"
                perc = (1 - change) * -100
                display = f"({perc:+.2f}%, {diff:+.2f})".rjust(17)
                display = click.style(display, fg=color)
            elif 1.01 <= change:
                if change < 1.05:
                    color = "yellow"
                else:
                    color = "red"
                perc = (change - 1) * 100
                display = f"({perc:+.2f}%, {diff:+.2f})".rjust(17)
                display = click.style(display, fg=color)
        key = click.style(key, fg="blue")
        echo(f"{key}: {value_d}{dev} {display}")

    if (
        minimal_relative_change is not None
        and max_abs_change < minimal_relative_change
    ):
        return used_compare_key
    if minimal_real_change is not None and max_abs_diff < minimal_real_change:
        return used_compare_key

    for one_line in lines:
        click.echo(one_line)

    return used_compare_key


def compare_results(
    path_file,
    json_file,
    dimension=None,
    ignored_dimensions=(),
    reference_key=None,
    value_type="median",
    metric_type="wall",
    minimal_relative_change=None,
    minimal_real_change=None,
):
    """Compare results together"""
    # XXX the API is awful
    paths = None
    all_results = []
    if path_file is None and json_file is None:
        paths = [p.rstrip("\n") for p in sys.stdin.readlines()]
    elif path_file is not None:
        assert json_file is None
        if hasattr(path_file, "readlines"):
            paths = [p.rstrip("\n") for p in path_file.readlines()]
        else:
            with open(path_file) as f:
                paths = [p.rstrip("\n") for p in f.readlines()]
    else:
        all_results = json.load(json_file)
    if paths is not None:
        for path in paths:
            all_results.append(get_data(path))

    results = dispatch_results(
        all_results,
        dimension,
        ignored_dimensions=ignored_dimensions,
    )
    unused_compare_key = display_all_results(
        results,
        reference_key=reference_key,
        value_type=value_type,
        metric_type=metric_type,
        minimal_relative_change=minimal_relative_change,
        minimal_real_change=minimal_real_change,
    )
    if unused_compare_key:
        click.secho(
            f"Warning: no results found with the dimension '{dimension}'",
            err=True,
            fg="yellow",
        )
