import os
import random
from . import (
    basics,
    errors,
)


def get_data_env(data_env_path):
    data_env_desc = basics.data_env_file(data_env_path)
    data = basics.get_data(data_env_desc)
    if data is None:
        raise FileNotFoundError(
            f"data environment {data_env_path} does not exist"
        )
    return DataEnvironment(data_env_path, data)


class DataEnvironment:
    def __init__(self, path, data):
        self._path = path
        self._data = data

    @property
    def path(self):
        return self._path

    @property
    def data_env_vars(self):
        return self._data["data-env-vars"].copy()

    def write(self):
        basics.write_data(basics.data_env_file(self._path), self._data)

    def get_benchmark_input_vars(self, key):
        full_key = f"bench-input-vars.{key}"
        value = basics.get_one_value(self._data, full_key)
        if value is None:
            raise errors.MissingDataEnvInputVar(key)
        return value

    def has_benchmark_input_vars(self, key: str) -> bool:
        """return True is the benchmark input variable is available"""
        try:
            self.get_benchmark_input_vars(key)
        except errors.MissingDataEnvInputVar:
            return False
        else:
            return True

    def set_benchmark_input_var(self, key, value):
        full_key = f"bench-input-vars.{key}"
        basics.set_one_value(self._data, full_key, value)

    def set_applied_brick_items(self, brick_name, items):
        key = f"applied-bricks.{brick_name}"
        basics.set_one_value(self._data, key, items)


def brick_apply_create_files(repo_path, brick_file, brick_temp_dir, params):
    data = (brick_temp_dir / params["files_list"]).read_bytes()
    list_of_files = data.split(b"\n")

    for fname in list_of_files:
        if not fname:
            continue
        new_file = repo_path / os.fsdecode(fname)
        new_file.parent.mkdir(parents=True, exist_ok=True)
        new_file.touch()


def brick_apply_modify_files(repo_path, brick_file, brick_temp_dir, params):
    data = (brick_temp_dir / params["files_list"]).read_bytes()
    list_of_files = data.split(b"\n")

    for fname in list_of_files:
        if not fname:
            continue
        file_path = repo_path / os.fsdecode(fname)
        if not file_path.exists():
            raise errors.InvalidBrick(
                f"file {file_path} to modify does not exist", brick_file
            )
        file_path.write_bytes(b"poulpe modified file %s" % file_path)


def brick_apply_delete_files(repo_path, brick_file, brick_temp_dir, params):
    data = (brick_temp_dir / params["files_list"]).read_bytes()
    list_of_files = data.split(b"\n")

    for fname in list_of_files:
        if not fname:
            continue
        file_path = repo_path / os.fsdecode(fname)
        file_path.unlink()


brick_appliers = {
    "poulpe::create-files": brick_apply_create_files,
    "poulpe::modify-files": brick_apply_modify_files,
    "poulpe::delete-files": brick_apply_delete_files,
}
