"""
A Poulpe Den is a directory with content structed in a standard way that
simplify the usage of Poulpe for benchmarking. It constains and manages the
basic building block of Poulpe
"""

from __future__ import annotations

import glob
import os
import re

from pathlib import Path
import sys

import click
from poulpe import basics
from poulpe import benchmarks
from poulpe import binenvs as bin_env_mod
from poulpe import dataenvs
from poulpe import filter


class PoulpeDen:
    def __init__(self, root: Path):
        self.root_dir = root

    @property
    def bin_dir(self):
        return self.root_dir / "bin"

    @property
    def bin_env_dir(self):
        return self.root_dir / "bin-envs"

    @property
    def data_env_dir(self):
        return self.root_dir / "data-envs"

    @property
    def benchmark_dir(self):
        return self.root_dir / "benchmarks"

    @property
    def results_dir(self):
        return self.root_dir / "results"

    @property
    def suites_dir(self):
        return self.root_dir / "suites"

    @property
    def source_repositories_dir(self):
        return self.root_dir / "repositories"

    @property
    def tooling_dir(self):
        return self.root_dir / "tooling"

    def find_data_envs(self, filters=()):
        """walk the data-envs directory in search of data-env"""
        data_envs = self._find_all_data_envs()
        result = []
        for d in data_envs:
            if filter.apply_filters(filters, d.data_env_vars):
                result.append(d)
        return result

    def _find_all_data_envs(self):
        den_data_envs = []

        # search for data_envs in the data_env directory
        candidates_stack = []
        for x in self.data_env_dir.iterdir():
            if x.is_dir() and not x.name.startswith("."):
                candidates_stack.append([MAX_DATA_ENVS_DEPTH, x])

        while candidates_stack:
            ttl, current = candidates_stack.pop()
            ttl -= 1
            # XXX have the data_env_dir function return Path object directly
            if Path(basics.data_env_file(current)).exists():
                den_data_envs.append(dataenvs.get_data_env(current))
            elif ttl >= 0:
                for x in current.iterdir():
                    if x.is_dir():
                        candidates_stack.append([ttl, x])
        den_data_envs.sort(key=lambda d: d.path)
        return den_data_envs

    def find_benchmarks(self, pattern=None):
        abs_files = set()
        root = self.benchmark_dir
        abs_root = str(root.resolve())

        lvl_1_glob = glob.escape(abs_root) + "/*/*.pbd"
        lvl_2_glob = glob.escape(abs_root) + "/*/*/*.pbd"
        # XXX the `root_dir` argument is not supported until python 3.10
        abs_files.update(glob.glob(lvl_1_glob, recursive=True))
        # bypass the usual top level symlink
        abs_files.update(glob.glob(lvl_2_glob, recursive=True))
        rel_files = [f[len(abs_root) + 1 :] for f in abs_files]
        files = [root / f for f in rel_files]
        bms = [(f, benchmarks.get_benchmark(f)) for f in files]
        pre_filter_bms = [(f, benchmarks.get_benchmark(f)) for f in files]
        if pattern is None:
            bms = pre_filter_bms
        else:
            target = filter.build_pattern(pattern)
            bms = []
            for f, bench in pre_filter_bms:
                if target.match(bench.name) is not None:
                    bms.append((f, bench))

        return sorted(bms, key=lambda x: x[1].name)

    def find_bin_setup(self) -> list[tuple[str, Path]]:
        all_bin_setup = []
        for suite in self.suites_dir.glob("./*/"):
            bin_setup = suite / "bin-env-setup/default.poulpe-bin-env-setup"
            if bin_setup.exists():
                # XXX we should have a class for them
                all_bin_setup.append((suite.name, bin_setup))
        all_bin_setup.sort()
        return all_bin_setup


def check(den: PoulpeDen):
    for display, attr, check_func in EXPECTED_DIRS:
        path = getattr(den, attr)
        if not path.exists():
            relpath = path.relative_to(den.root_dir)
            print(display, f"[MISSING] (./{relpath}/)")
        else:
            print(display, "[OKAY]")
            if check_func is not None:
                check_func(den)


MAX_DATA_ENVS_DEPTH = 5


def _check_data_envs(den):
    den_data_envs = den.find_data_envs()
    print(f"  available: {len(den_data_envs)}")


def _check_benchmarks(den):
    den_benchmarks = den.find_benchmarks()
    print(f"  available: {len(den_benchmarks)}")


def _check_suites(den):
    bin_setup_scripts = den.find_bin_setup()
    print(f"  available bin setup: {len(bin_setup_scripts)}")


# list of (<print name>, <attribute name>) tuple
EXPECTED_DIRS = [
    ("script and tools:     ", "bin_dir", None),
    ("binary environments:  ", "bin_env_dir", None),
    ("data environments:    ", "data_env_dir", _check_data_envs),
    ("benchmarks:           ", "benchmark_dir", _check_benchmarks),
    ("results:              ", "results_dir", None),
    ("poulpe suites:        ", "suites_dir", _check_suites),
    ("source repositories:  ", "source_repositories_dir", None),
    ("tooling and utilities:", "tooling_dir", None),
]


def list_benchmarks(den, pattern=None, list_variants=False, show_path=False):
    for bench_path, bench in den.find_benchmarks(pattern=pattern):
        print(bench.name)
        if show_path:
            print(f" {bench_path}")
        if list_variants:
            for dimension, variants in sorted(bench.all_dimensions.items()):
                print(f"  * {dimension}:")
                for v, values in variants.items():
                    extra = ""
                    if values.get("default", False):
                        extra = " (default)"
                    print(f"    - {v}" + extra)


def list_bin_env_setup(den):
    for suite_name, bin_env_setup_path in den.find_bin_setup():
        bin_env_setup = bin_env_mod.BinEnvSetup(bin_env_setup_path)
        print(f"{suite_name}:")
        parameters = list(bin_env_setup.get_parameters().items())
        max_name_length = max(len(p[0]) for p in parameters)
        for name, details in parameters:
            padded_name = name.ljust(max_name_length)
            print(f'  * {padded_name} {details["doc"]}')
            if "default" not in details:
                print("    - REQUIRED")
            else:
                default_value = details["default"]
                print(f'    - Optional (default: "{default_value}")')
            arg_type = details["type"]
            print(f"    - type: {arg_type}")
            if arg_type == "choice":
                choices = list(details["choices"].items())
                max_choice_length = max(len(c[0]) for c in choices) + 1
                for c, d in details["choices"].items():
                    c = c.ljust(max_choice_length)
                    print(f"      ~ {c} {d}")


def interactive_bin_env_spec(den: "PoulpeDen"):
    # avoid circular import
    from poulpe.scheduling import BinEnvSpec

    suites: dict[str, bin_env_mod.BinEnvSetup.ParameterT] = {}
    for suite_name, bin_env_setup_path in den.find_bin_setup():
        bin_env_setup = bin_env_mod.BinEnvSetup(bin_env_setup_path)
        suites[suite_name] = bin_env_setup.get_parameters()

    if not suites:
        click.secho("Could not find any suites, aborting", file=sys.stderr)
        exit(1)

    click.echo(
        f"Note: you can find all setup parameters by running"
        f"`poulpe den-list-bin-env-setups`\n"
    )

    if len(suites) == 1:
        suite = list(suites)[0]
        click.secho(f"Found a single suite '{suite}', picking it", fg="yellow")
    else:
        suite = click.prompt(
            "Which suite would you like to test?",
            type=click.Choice(list(suites.keys())),
            default=list(suites)[0],
        )

    params = suites[suite]
    bin_env_args = {}
    total = len(params)

    for idx, (name, info) in enumerate(params.items(), start=1):
        default = info.get("default")
        click.secho("---", fg="blue")
        optional = "(optional)" if default is not None else "(required)"
        click.secho(f"{name} {optional} [{idx}/{total}]", fg="blue")
        if "doc" in info:
            click.secho(info["doc"].capitalize(), fg="blue")
        if info["type"] == "choice":
            choices = info["choices"]
            click.secho("Choices available:", fg="blue")
            for choice, helptext in choices.items():
                click.secho(f"  - {choice}: {helptext}", fg="blue")
            value = click.prompt(
                f"?", type=click.Choice(choices), default=default
            )
        elif re.match("^<.*>$", info["type"]):
            # string literal
            click.secho(f"Expected type: {info['type']}", fg="blue")
            value = click.prompt(f"?", default=default)
        else:
            raise ValueError(f"unsupported parameter type {info['type']}")

        if value != default:
            bin_env_args[name] = value

    spec = BinEnvSpec(suite, bin_env_args)
    as_string = spec.task_bin_spec()

    # Check that it's valid round-trip
    round_trip = BinEnvSpec.from_piece(as_string).task_bin_spec()
    assert round_trip == as_string, (round_trip, as_string)

    click.echo("---")
    click.secho("Successfully generated the bin-env spec:", fg="green")
    click.echo(spec.task_bin_spec())


def list_data_env(den, list_vars=False, filters=()):
    for data_env in den.find_data_envs(filters=filters):
        print(f"{data_env.path}")
        if list_vars:
            basics.show(data_env.data_env_vars, print, indent="  ")
