from __future__ import annotations

import functools
import os
import pathlib
import psutil
import time
import shutil
import subprocess
import sys

import click

from . import (
    basics as poulpe,
    benchmarks as bench_mod,
    dataenvs as data_mod,
    binenvs as bin_mod,
    errors,
)

BTRFS_DEBUG = False

RETURN_CODE_SUCCESS = 0
RETURN_CODE_EXEC_ISSUE = 32
RETURN_CODE_SETUP_ISSUE = 64


def _file_system_for(path):
    """gather file system information for `path`"""
    best = None
    prefix = str(path.resolve())
    for p in psutil.disk_partitions(all):
        if prefix.startswith(p.mountpoint):
            if best is None:
                best = p
            elif len(best.mountpoint) < len(p.mountpoint):
                best = p
    if best is None:
        msg = "cannot determine file system: %s"
        raise RuntimeError(msg % prefix)
    return best


NO_SNAPSHOT = 1
COULD_SNAPSHOT_BUT_NOT_SUBVOLUME = 3
COULD_SNAPSHOT_BUT_NOT_DELETE = 5
CAN_SNAPSHOT = 4

COPY_AS_SNAPSHOT = {COULD_SNAPSHOT_BUT_NOT_DELETE, CAN_SNAPSHOT}


def _may_snapshot(path):
    """return the situation regarding snapshot usage"""
    path = path.resolve()
    fs = _file_system_for(path)
    options = fs.opts.split(",")
    if fs.fstype != "btrfs":
        return NO_SNAPSHOT
    elif os.stat(path).st_ino != 256:
        return COULD_SNAPSHOT_BUT_NOT_SUBVOLUME
    elif "user_subvol_rm_allowed" not in options:
        return COULD_SNAPSHOT_BUT_NOT_DELETE
    elif not os.path.ismount(path):
        return COULD_SNAPSHOT_BUT_NOT_SUBVOLUME
    else:
        return CAN_SNAPSHOT


def _btrfs_sub_delete(path):
    """delete a btrsf sub volume"""
    cp_cmd = [
        "btrfs",
        "--quiet",
        "sub",
        "delete",
        "--commit-after",
        path,
    ]
    subprocess.check_call(cp_cmd)


def _btrfs_sub_snap(src, dst):
    """do a btrfs snapshot from `src` to `dst`"""
    cp_cmd = [
        "btrfs",
        "--quiet",
        "sub",
        "snap",
        src,
        dst,
    ]
    subprocess.check_call(cp_cmd)


def copy_data_env(reference, tmp):
    """ensure we have a clean copy of the `reference` data-env at `tmp` path

    This may use a btrfs snapshot.
    """
    cleanup_data_env(tmp)  # remove potential remains of previous runs

    do_snapshot = NO_SNAPSHOT

    # IMPORTANT NOTE: here assume we always copy the data env on the same file
    # system as the source. It is important regarless of using btrfs so we can
    # assume this will remains the case in the future.
    do_snapshot = _may_snapshot(reference)
    if BTRFS_DEBUG and do_snapshot == COULD_SNAPSHOT_BUT_NOT_SUBVOLUME:
        m = "skipping btrfs snapshot: data env not a subvolume"
        print(m, file=sys.stderr)

    start = time.monotonic()
    if do_snapshot in COPY_AS_SNAPSHOT:
        _btrfs_sub_snap(reference, tmp)
    else:
        # we also use btrfs snapshot when applicable
        cp_cmd = [
            "cp",
            "--archive",
            "--reflink=auto",
            reference,
            tmp,
        ]
        subprocess.check_call(cp_cmd)
    end = time.monotonic()
    duration = end - start
    if BTRFS_DEBUG:
        m = f"data env duplication took {duration:.3f} seconds"
        print(m, file=sys.stderr)


def cleanup_data_env(tmp_data):
    """remove a temporary data-env"""
    if tmp_data is None or not tmp_data.exists():
        return
    start = time.monotonic()
    do_snapshot = _may_snapshot(tmp_data)
    if do_snapshot == CAN_SNAPSHOT:  # this a btrf snapshot
        _btrfs_sub_delete(tmp_data)
    else:
        if BTRFS_DEBUG and do_snapshot == COULD_SNAPSHOT_BUT_NOT_DELETE:
            m = "skipping btrfs based cleanup: missing user_subvol_rm_allowed option"
            print(m, file=sys.stderr)
        shutil.rmtree(tmp_data, ignore_errors=True)
    end = time.monotonic()
    duration = end - start
    if BTRFS_DEBUG:
        m = f"data env cleanup took {duration:.3f} seconds"
        print(m, file=sys.stderr)


def run_one_core(
    bin_env_path: pathlib.Path,
    data_env_path: pathlib.Path,
    benchmark_path: str,
    debug=False,
):
    result_data: dict[str, dict] = {}
    result_data["run"] = {}
    result_data["run"]["timestamp"] = time.time()

    # read the benchmark data
    benchmark = bench_mod.get_benchmark(benchmark_path)
    assert benchmark is not None, benchmark_path

    result_data["benchmark"] = {}
    result_data["benchmark"]["name"] = benchmark.name
    variants = benchmark.selected_variants
    if variants:
        result_data["benchmark"]["variants"] = variants

    # gather info about the binary environment
    bin_env = bin_mod.get_binary_env(bin_env_path)
    result_data["bin-env-vars"] = bin_env.bin_env_vars

    run_in_copy = False
    data_env_path = pathlib.Path(data_env_path)
    if benchmark.get_var("run.copy-data-env", True):
        run_in_copy = True
    elif _may_snapshot(data_env_path) in COPY_AS_SNAPSHOT:
        run_in_copy = True
    tmp_data = None

    if run_in_copy:
        tmp_dir = data_env_path.parent / ".tmp/"
        os.makedirs(tmp_dir, exist_ok=True)
        tmp_data = tmp_dir / data_env_path.name
        copy_data_env(data_env_path, tmp_data)
        data_env_path = tmp_data

    # gather info about the data environment
    data_env_data = data_mod.get_data_env(data_env_path)
    result_data["data-env-vars"] = data_env_data.data_env_vars

    try:
        bench_result = benchmark.run_one(
            bin_env_path, data_env_path, debug=debug
        )
        result_data["result"] = bench_result
    finally:
        cleanup_data_env(tmp_data)
        duration = time.time() - result_data["run"]["timestamp"]
        result_data["run"]["duration"] = duration
    return result_data


def run_one(bin_env_path, data_env_path, benchmark_path, result, debug=False):
    try:
        result_data = run_one_core(
            bin_env_path,
            data_env_path,
            benchmark_path,
            debug=debug,
        )
    except errors.MissingDataEnvInputVar as exc:
        # TODO it's not the lib's responsibility to write to stderr
        err = functools.partial(click.secho, err=True, fg="red")
        msg = (
            f"ABORT: data-environment is missing an input-value "
            f"definition for: {exc.args[0]}"
        )
        err(msg)
        return RETURN_CODE_SETUP_ISSUE
    except errors.BenchmarkRunFailure as exc:
        # TODO it's not the lib's responsibility to write to stderr
        err = functools.partial(click.secho, err=True, fg="red")
        err("ERROR: running benchmark failed:")
        err("ERROR:   benchmark details:")
        err(f"ERROR:     bin-env:   {bin_env_path}")
        err(f"ERROR:     data-env:  {data_env_path}")
        err(f"ERROR:     benchmark: {benchmark_path}")
        err("ERROR:   run details:")
        if exc.cwd is not None:
            err(f"ERROR:     cwd:       {exc.cwd}")
        if exc.command is not None:
            err(f"ERROR:     command:   {repr(exc.command)}")
        if exc.return_code is not None:
            err(f"ERROR:     return:    {exc.return_code}")
        if exc.stdout is not None:
            err("ERROR:     std-out:")
            for line in exc.stdout.splitlines():
                err(f"ERROR:       {line.decode('utf8')}")
        if exc.stderr is not None:
            err("ERROR:     std-err:")
            for line in exc.stderr.splitlines():
                err(f"ERROR:       {line.decode('utf8')}")
        if exc.message is not None:
            err("ERROR:     poulpe:")
            for line in exc.message.splitlines():
                err(f"ERROR:       {line}")
        return RETURN_CODE_EXEC_ISSUE
    else:
        poulpe.write_data(result, result_data)
        return RETURN_CODE_SUCCESS
