import datetime
import hashlib
import os
import re
import shutil
import subprocess
import sys
import tempfile
import time

from pathlib import Path

import click

from poulpe import (
    den as den_mod,
    binenvs as bin_env_mod,
    errors,
)
from poulpe.runner import (
    RETURN_CODE_EXEC_ISSUE,
    RETURN_CODE_SETUP_ISSUE,
    RETURN_CODE_SUCCESS,
)
from poulpe.utils import escape_string, split_unescape


def generate_tasks(
    den,
    bin_setup_specs,
    all_variants=False,
    data_env_filters=(),
    benchmark_name_pattern=(),
    benchmark_variant_filters=(),
):
    binaries = list_binary_variants(den, bin_setup_specs)
    data_envs = den.find_data_envs(filters=data_env_filters)
    benchmarks = list(
        list_benchmarks_variants(
            den,
            all_variants=all_variants,
            name_pattern=benchmark_name_pattern,
            variant_filters=benchmark_variant_filters,
        )
    )
    for bin in binaries:
        bin_spec = str(bin.task_bin_spec())
        for d in data_envs:
            data_path = str(d.path.relative_to(den.data_env_dir))
            for b_spec, bench in benchmarks:
                if bench.compatible_with(bin_env=bin, data_env=d):
                    yield ("RUN", bin_spec, data_path, b_spec)


def list_binary_variants(den, bin_setup_specs):
    variants = []

    target_bins = [BinEnvSpec.from_piece(s) for s in bin_setup_specs]

    available = {}
    for suite_name, bin_env_setup_path in den.find_bin_setup():
        available[suite_name] = bin_env_mod.BinEnvSetup(bin_env_setup_path)

    for target in target_bins:
        setup = available[target.suite_name]
        parameters = setup.get_parameters()
        for name, info in parameters.items():
            if info.get("default") is None and name not in target.arguments:
                raise errors.MissingDataEnvInputVar(
                    f"bin-env setup '{setup._path}' expects a value for the '{name}' parameter"
                )

        # TODO create the missing combinations?
        variants.append(target)

    return variants


def list_benchmarks_variants(
    den,
    all_variants=True,
    name_pattern=None,
    variant_filters=(),
):
    benchmarks = den.find_benchmarks()
    for bench_path, bench in benchmarks:
        if name_pattern is not None and name_pattern.match(bench.name) is None:
            continue
        rel_path = bench_path.relative_to(den.benchmark_dir)
        all_dimensions = sorted(bench.all_dimensions.items())
        if not all_dimensions:
            yield (str(rel_path), bench)
        elif not (variant_filters or all_variants):
            yield (str(rel_path), bench)
        else:
            for kv in combine_dimensions(
                all_dimensions,
                variant_filters,
                all_variants=all_variants,
            ):
                if kv:
                    v_spec = _format_variants(kv)
                    p = f"{rel_path}:{v_spec}"
                else:
                    p = str(rel_path)
                yield (p, bench.new_with_variants(kv))


def _format_variants(variants):
    items = sorted(variants.items())
    return ":".join("%s=%s" % i for i in items)


def combine_dimensions(
    all_dimensions,
    variant_filters=(),
    all_variants=False,
):
    if len(all_dimensions) <= 0:
        assert False
    head = all_dimensions[0]
    tail = all_dimensions[1:]

    filters = {}
    for path, pattern in variant_filters:
        assert len(path) == 1
        filters[path[0]] = pattern

    this_level = []
    f = filters.get(head[0])
    if all_variants or f is not None:
        for v in head[1]:
            if f is not None:
                if f.match(v) is None:
                    continue
            this_level.append({head[0]: v})
    elif not all_variants:
        this_level.append({})

    if not tail:
        yield from this_level
    else:
        for kv in this_level:
            for sub in combine_dimensions(
                tail,
                variant_filters,
                all_variants,
            ):
                n = kv.copy()
                if sub:
                    n.update(sub)
                yield n


class BinEnvSpec:
    @classmethod
    def from_piece(cls, piece: str):
        suite_name, *bin_env_vars = split_unescape(piece, ":")

        as_dict = {}
        for d in bin_env_vars:
            try:
                k, v = split_unescape(d, "=")
            except ValueError as e:
                raise ValueError(f"{d}: {e}") from e
            as_dict[k] = v

        return cls(
            suite_name=suite_name,
            arguments=as_dict,
        )

    def __init__(self, suite_name: str, arguments=None):
        self.suite_name = suite_name
        if arguments is None:
            arguments = {}
        self.arguments = arguments

    def to_string_id(self):
        """return a string one can use as a unique directory name"""
        # XXX might need the Den to get more data about the setup script and
        # fill the default value of unspecified parameters
        pieces = [self.suite_name]
        for key in sorted(self.arguments):
            value = self.arguments[key]
            pieces.append(f"{key}={value}")
        return hashlib.sha256(":".join(pieces).encode()).hexdigest()

    def task_bin_spec(self):
        pieces = [self.suite_name]
        for key in sorted(self.arguments):
            value = escape_string(self.arguments[key], ":=")
            pieces.append(f"{key}={value}")
        return ":".join(pieces)


class DataEnvSpec:
    @classmethod
    def from_piece(cls, piece: str):
        return cls(data_env_path=piece)

    def __init__(self, data_env_path: str):
        self.data_env_path = data_env_path

    def to_string_id(self):
        return self.data_env_path


class BenchmarkSpec:
    @classmethod
    def from_piece(cls, piece: str):
        benchmark_path, *bench_variants = split_unescape(piece, ":")

        as_dict = {}
        for d in bench_variants:
            k, v = split_unescape(d, "=")
            as_dict[k] = v
        return cls(
            benchmark_path=benchmark_path,
            arguments=as_dict,
        )

    def __init__(self, benchmark_path: str, arguments=None):
        self.benchmark_path = benchmark_path
        if arguments is None:
            arguments = {}
        self.arguments = arguments

    def to_string_id(self):
        """return a string one can use as a unique directory name"""
        # XXX might need the Den to get more data about the setup script and
        # fill the default value of unspecified parameters
        pieces = [self.benchmark_path]
        for key in sorted(self.arguments):
            value = escape_string(self.arguments[key], ":=")
            pieces.append(f"{key}={value}")
        return hashlib.sha256(":".join(pieces).encode()).hexdigest()

    def benchmark_run_spec(self, den):
        pieces = [str(den.benchmark_dir / self.benchmark_path)]
        for key in sorted(self.arguments):
            value = escape_string(self.arguments[key], ":=")
            pieces.append(f"{key}={value}")
        return ":".join(pieces)

    def task_spec(self):
        pieces = [self.benchmark_path]
        for key in sorted(self.arguments):
            value = escape_string(self.arguments[key], ":=")
            pieces.append(f"{key}={value}")
        return ":".join(pieces)


class RunTask:
    type = "run"

    TASK_KEYWORD = "RUN"

    @classmethod
    def from_line(
        cls,
        line,
    ):
        bin_env_piece, data_env_piece, benchmark_piece = line.split()
        return cls(
            bin_env_spec=BinEnvSpec.from_piece(bin_env_piece),
            data_env_spec=DataEnvSpec.from_piece(data_env_piece),
            benchmark_spec=BenchmarkSpec.from_piece(benchmark_piece),
        )

    def __init__(
        self,
        bin_env_spec: BinEnvSpec,
        data_env_spec: DataEnvSpec,
        benchmark_spec: BenchmarkSpec,
    ):
        self.bin_env_spec = bin_env_spec
        self.data_env_spec = data_env_spec
        self.benchmark_spec = benchmark_spec

    def result_id(self):
        """an arbitrary unique string that encode the task parameter"""

        result_desc = " ".join(
            [
                self.bin_env_spec.to_string_id(),
                self.data_env_spec.to_string_id(),
                self.benchmark_spec.to_string_id(),
            ]
        )
        return hashlib.sha256(result_desc.encode("utf8")).hexdigest()

    def as_log_line(self):
        bin = click.style(self.bin_env_spec.task_bin_spec())
        data = click.style(self.data_env_spec.data_env_path, underline=True)
        benchmark_spec = click.style(self.benchmark_spec.task_spec())
        return f"{bin} {data} {benchmark_spec}"


ALL_TASK_CLASSES = [
    RunTask,
]

TASK_CLASS_MAPPING = dict((a.TASK_KEYWORD, a) for a in ALL_TASK_CLASSES)


def get_run_tasks(den: den_mod.PoulpeDen, schedule_path: Path):
    tasks = _get_all_tasks(den, schedule_path)
    return [a for a in tasks if a.type == RunTask.type]


def _get_all_tasks(den: den_mod.PoulpeDen, schedule_path: Path):
    raw_tasks = schedule_path.read_text()
    tmp_tasks = [a.strip() for a in raw_tasks.splitlines()]
    tasks = [a.split(maxsplit=1) for a in tmp_tasks if a]
    return [
        TASK_CLASS_MAPPING[a[0]].from_line(a[1])
        for a in tasks
        if a[0] in TASK_CLASS_MAPPING and a[0] != "#"
    ]


def run_multiple_tasks(den, tasks, report_path=None, lazy=False):
    total = len(tasks)
    successes = 0
    setup_issues = 0
    run_issues = 0
    unknown_failures = 0

    for idx, current_task in enumerate(tasks, start=1):
        ret = run_one_task(
            den,
            current_task,
            progress_info=(idx, total),
            report_path=report_path,
            lazy=lazy,
        )
        if ret == RETURN_CODE_SUCCESS:
            successes += 1
        elif ret == RETURN_CODE_SETUP_ISSUE:
            setup_issues += 1
        elif ret == RETURN_CODE_EXEC_ISSUE:
            run_issues += 1
        else:
            unknown_failures += 1
    return (successes, setup_issues, run_issues, unknown_failures)


def run_one_task(
    den: den_mod.PoulpeDen,
    task: RunTask,
    progress_info=None,
    report_path=None,
    lazy=False,
) -> int:
    bin_env = ensure_bin_env(den, task.bin_env_spec)
    if bin_env is None:
        print("could not get running bin-env", file=sys.stderr)
        return -1
    result = den.results_dir / f"{task.result_id()}.pbr"

    data_env = den.data_env_dir / task.data_env_spec.data_env_path
    benchmark = task.benchmark_spec.benchmark_run_spec(den)

    log_line = task.as_log_line()
    u_log_line = click.unstyle(log_line)

    if lazy and result.exists():
        pieces = [click.style("skipping", fg="blue"), log_line]
        if progress_info is not None:
            pieces.insert(0, click.style("[%d/%d]" % progress_info, fg="cyan"))
        click.echo(" ".join(pieces))
        click.secho("(a result already exists for this task)", fg="blue")
        if report_path is not None:
            start = datetime.datetime.now()
            m_start = time.monotonic()
            # XXX lazily assuming we do not need locking here as running
            # multiple benchmark at the same time seems silly.
            with open(report_path, "a") as report:
                print(start, "### skipping:", u_log_line, file=report)
                print(start, "    result:  ", str(result), file=report)
        return 0

    pieces = [click.style("running", fg="green"), log_line]
    if progress_info is not None:
        pieces.insert(0, click.style("[%d/%d]" % progress_info, fg="cyan"))
    click.echo(" ".join(pieces))

    if report_path is not None:
        start = datetime.datetime.now()
        m_start = time.monotonic()
        # XXX lazily assuming we do not need locking here as running multiple
        # benchmark at the same time seems silly.
        with open(report_path, "a") as report:
            print(start, "### starting:", u_log_line, file=report)
    cmd = [
        "poulpe",
        "run-util",
        str(bin_env),
        str(data_env),
        str(benchmark),
        str(result),
    ]
    ret = subprocess.call(cmd)
    if report_path is not None:
        end = datetime.datetime.now()
        m_end = time.monotonic()
        duration = m_end - m_start
        # XXX lazily assuming we do not need locking here as running multiple
        # benchmark at the same time seems silly.
        with open(report_path, "a") as report:
            print(end, "    duration: %d seconds" % duration, file=report)
            print(end, "    return:   %d" % ret, file=report)
            print(end, "    result:  ", str(result), file=report)

    return ret


# piece doing automated setup of bin-env

MAX_BIN_ENV = 50


def is_ready(bin_env):
    """wtf, need better"""
    desc = bin_env / "bin-env.poulpe"
    if not desc.exists():
        return False
    return "ready = true" in desc.read_text()


def ensure_bin_env(den: den_mod.PoulpeDen, spec: BinEnvSpec):
    bin_env_dir = den.bin_env_dir

    # TODO parametrize which script somehow
    suite_name = spec.suite_name
    setup_script = (
        den.suites_dir
        / suite_name
        / "bin-env-setup"
        / "default.poulpe-bin-env-setup"
    )
    bin_env_setup = bin_env_mod.BinEnvSetup(setup_script)
    parameters = bin_env_setup.get_parameters()
    cli_args = {}
    env_args = {}
    for name, value in spec.arguments.items():
        if name not in parameters:
            raise bin_env_mod.ParametersError(
                f"unexpected parameter '{name}' for '{setup_script}'"
            )
        arg_mode = parameters[name]["arg-mode"]
        assert arg_mode in ("CLI", "ENV"), arg_mode
        if arg_mode == "CLI":
            cli_args[name] = value
        else:
            env_args[name] = value

    env = os.environ.copy()
    env.update(env_args)
    cli_args_list = []
    if cli_args:
        for name, value in cli_args.items():
            cli_args_list.append(f"--{name}")
            cli_args_list.append(value)

    if suite_name == "hg":
        # Backwards compatibility for hg suite
        # XXX make the mercurial script do this, I can't be bothered to
        # XXX edit a giant shell script right now
        target = bin_env_dir / spec.to_string_id()
    else:
        tmp_target = tempfile.mkdtemp()

        # TODO make a "poulpe resolve-bin-env-spec"?
        cmd = [setup_script.resolve(), "--get-unique-id"] + cli_args_list

        # Get the actual unique bin-env ID that only the suite can resolve
        # (because of its dependencies, etc.)
        res = subprocess.run(
            cmd,
            cwd=tmp_target,
            env=env,
            capture_output=True,
        )
        try:
            res.check_returncode()
            output = res.stdout.decode()
            match = re.match("UNIQUE_ID: (.*)", output)
            assert match is not None, output
            unique_id = match.group(1)
        except Exception:
            raise
        finally:
            shutil.rmtree(tmp_target, ignore_errors=True)

        target = bin_env_dir / unique_id
    # check the already available bin-env
    all_existing = [p for p in bin_env_dir.iterdir()]
    all_existing.sort(key=lambda x: x.stat().st_mtime)

    assert MAX_BIN_ENV > 1
    if len(all_existing) > MAX_BIN_ENV:
        for delete_me in all_existing:
            if delete_me != target:
                break
        print(f"Bin-envs {MAX_BIN_ENV} limit, deleting: {delete_me}")
        shutil.rmtree(delete_me)

    # that check is too simple, we need to make sure this is a valid bin-env
    # too
    if target.exists() and not is_ready(target):
        shutil.rmtree(target)
    if not target.exists():
        os.makedirs(target, exist_ok=True)

        cmd = ["poulpe", "bin-env-util"]
        cmd.extend(
            [
                "setup-one",
                target,
                setup_script,
                "--" if cli_args_list else "",
            ]
            + cli_args_list
        )
        # XXX check the result of this at some point…
        subprocess.run(cmd, env=env)

    if is_ready(target):
        return target
    else:
        shutil.rmtree(target)
        return None
