# Useful common code
import re
from typing import Iterable, List


def escape_string(s: str, chars: str, escape: str = "\\") -> str:
    """Escape all unescaped characters in `chars` with `escape`"""
    for delim in chars:
        delim = re.escape(delim)
        s = re.sub(rf"([^{delim}]){delim}", rf"\g<1>{escape}{delim}", s)
    return s


def join_escaped(to_join: Iterable[str], chars: str, escape: str = "\\") -> str:
    """Concatenate all strings in `to_join`, separating them with `chars`,
    after escaping each char in `chars`."""
    return chars.join(escape_string(s, chars, escape) for s in to_join)


def split_unescape(s: str, delim: str, escape: str = "\\") -> List[str]:
    """Split a string along `delim`, only if not escaped by `escape`

    >>> split_unescape('foo,bar', ',')
    ['foo', 'bar']
    >>> split_unescape('foo,bar', ',')
    ['foo', 'bar']
    >>> split_unescape('foo$,ba$r', ',', '$')
    ['foo,ba$r']
    >>> split_unescape('foo$$,bar', ',', '$')
    ['foo$', 'bar']
    >>> split_unescape('foo$', ',', '$')
    ['foo$']
    """
    ret = []
    current = []
    itr = iter(s)
    for ch in itr:
        if ch == escape:
            try:
                # skip the next character; it has been escaped!
                next_char = next(itr)
                if next_char == delim:
                    current.append(next_char)
                else:
                    if next_char != escape:
                        current.append(escape)
                    current.append(next_char)
            except StopIteration:
                current.append(escape)
        elif ch == delim:
            # split! (add current to the list and reset it)
            ret.append("".join(current))
            current = []
        else:
            current.append(ch)
    ret.append("".join(current))
    return ret
