use std::collections::HashMap;
use std::error::Error;
use std::path::{Path, PathBuf};

use crate::benchmark::{Benchmark, BenchmarkMethod};
use crate::binary_environment::BinaryEnvironment;
use crate::data_environment::DataEnvironment;
use crate::result::BenchmarkResult;
use crate::suite::{gather_suite, Suite};

pub struct Den {
    pub root: PathBuf,
    pub data_envs: Vec<DataEnvironment>,
    pub bin_envs: Vec<BinaryEnvironment>,
    pub suites: Vec<Suite>,
    pub benchmarks: Vec<Benchmark>,
    pub results: Vec<BenchmarkResult>,
}

impl Den {
    pub fn new_from_disk(den_root: &Path) -> Result<Self, Box<dyn Error>> {
        let data_envs = den_root
            .join("data-envs")
            .read_dir()?
            .filter_map(|entry| {
                let entry = entry.ok()?;
                let stat = std::fs::metadata(entry.path()).ok()?;
                if !stat.is_dir() {
                    return None;
                }
                let name = entry.file_name().into_string().ok()?;
                let root = entry.path();
                let metadata = toml::from_str(
                    &std::fs::read_to_string(root.join("data-env.poulpe"))
                        .ok()?,
                )
                .ok()?;
                Some(DataEnvironment {
                    root,
                    name,
                    metadata,
                })
            })
            .collect();
        let bin_envs = den_root
            .join("bin-envs")
            .read_dir()?
            .filter_map(|entry| {
                let entry = entry.ok()?;
                let stat = std::fs::metadata(entry.path()).ok()?;
                if !stat.is_dir() {
                    return None;
                }
                let root = entry.path();
                let metadata: toml::Value = toml::from_str(
                    &std::fs::read_to_string(root.join("bin-env.poulpe"))
                        .ok()?,
                )
                .ok()?;
                let ready = metadata
                    .get("ready")
                    .and_then(|v| v.clone().try_into().ok())
                    .unwrap_or(false);
                Some(BinaryEnvironment {
                    root,
                    spec: "TODO".to_string(),
                    metadata,
                    is_ready: ready,
                })
            })
            .collect();
        let mut benchmarks = vec![];
        let suites = den_root
            .join("suites")
            .read_dir()?
            .filter_map(|entry| gather_suite(den_root, &mut benchmarks, entry))
            .collect();

        benchmarks.sort_by_key(|a| a.qualified_name());

        let results = walkdir::WalkDir::new(den_root.join("results"))
            .follow_links(true)
            .into_iter()
            .filter_entry(|entry| match entry.path().extension() {
                None => true,
                Some(ext) => ext == "pbr",
            })
            .filter_map(|result| {
                let dir_entry = result.ok()?;
                if dir_entry.file_type().is_dir() {
                    return None;
                }
                let data: toml::Value = toml::from_str(
                    &std::fs::read_to_string(dir_entry.path()).ok()?,
                )
                .ok()?;

                Some(BenchmarkResult {
                    path: dir_entry.into_path(),
                    contents: data.clone().try_into().ok()?,
                    data,
                })
            })
            .collect();

        Ok(Den {
            root: den_root.to_owned(),
            data_envs,
            bin_envs,
            suites,
            benchmarks,
            results,
        })
    }

    pub fn find_suite_by_name<'a>(&'a self, name: &str) -> Option<&'a Suite> {
        self.suites.iter().find(|&suite| suite.name == name)
    }

    /// Returns all [`DataEnvironment`] compatible with this [`Benchmark`]
    /// given its `parameters`.
    pub fn compatible_data_envs(
        &self,
        benchmark: &Benchmark,
        parameters: &HashMap<String, String>,
    ) -> Vec<&DataEnvironment> {
        // TODO constraints?

        fn check_input_var(
            data_env: &DataEnvironment,
            value: &str,
        ) -> Option<()> {
            if let Some(value) = value.strip_prefix("DATA-VARS:") {
                data_env.get_bench_input_var(value)?;
            };
            Some(())
        }

        self.data_envs
            .iter()
            .filter_map(|data_env| {
                for value in benchmark.environment.values() {
                    check_input_var(data_env, value)?;
                }

                match &benchmark.method {
                    BenchmarkMethod::SimpleCommand(simple) => {
                        if let Some(variants) = &simple.variants {
                            for (parameter, value) in parameters {
                                let case = variants
                                    .dimensions
                                    .get(parameter)
                                    .expect("dimension should exist")
                                    .get(value)
                                    .expect("dimension value should be valid");
                                for value in case.environment.values() {
                                    check_input_var(data_env, value)?;
                                }
                                if let Some(cwd) = &case.cwd {
                                    check_input_var(data_env, cwd)?;
                                }
                                // TODO check more places?
                            }
                        }
                    }
                    BenchmarkMethod::MercurialPerfExtension(perf) => {
                        if let Some(variants) = &perf.variants {
                            for (parameter, value) in parameters {
                                let case = variants
                                    .dimensions
                                    .get(parameter)
                                    .expect("dimension should exist")
                                    .get(value)
                                    .expect("dimension value should be valid");
                                if let Some(args) = &case.args {
                                    for value in args {
                                        check_input_var(data_env, value)?;
                                    }
                                }
                                if let Some(cwd) = &case.cwd {
                                    check_input_var(data_env, cwd)?;
                                }
                                // TODO check more places?
                            }
                        }
                    }
                    BenchmarkMethod::Unknown => {}
                }

                Some(data_env)
            })
            .collect()
    }
}
