use std::{
    error::Error,
    os::unix::ffi::OsStrExt,
    path::{Path, PathBuf},
};

use crate::benchmark::{Benchmark, BenchmarkMethod};
use crate::binary_environment::Setup as BinEnvSetup;

pub struct Suite {
    pub name: String,
    pub data_env_setups: Vec<PathBuf>,
    pub bin_env_setups: Vec<BinEnvSetup>,
}

impl Suite {
    fn new(
        den_root: &Path,
        benchmarks: &mut Vec<Benchmark>,
        root: PathBuf,
        file_name: std::ffi::OsString,
    ) -> Result<Self, Box<dyn Error>> {
        let suite_name = file_name.to_str().ok_or("non UTF8 suite name")?;

        // TODO walkdir
        let data_env_setups =
            if let Ok(entries) = root.join("data-env-setup").read_dir() {
                entries
                    .filter_map(|entry| {
                        let entry = entry.ok()?;
                        let stat = entry.metadata().ok()?;
                        if !stat.is_file() {
                            return None;
                        }
                        let path = entry.path();
                        Some(path)
                    })
                    .collect()
            } else {
                vec![]
            };

        // TODO walkdir
        let bin_env_setup_paths =
            if let Ok(entries) = root.join("bin-env-setup").read_dir() {
                entries
                    .filter_map(|entry| {
                        let entry = entry.ok()?;
                        let stat = entry.metadata().ok()?;
                        if !stat.is_file() {
                            return None;
                        }
                        let path = entry.path();
                        Some(path)
                    })
                    .collect()
            } else {
                vec![]
            };

        // TODO walkdir
        if let Ok(entries) = root.join("benchmarks").read_dir() {
            benchmarks.extend(entries.filter_map(|entry| {
                let entry = entry.ok()?;
                let stat = std::fs::metadata(entry.path()).ok()?;

                if !stat.is_file() {
                    return None;
                }

                let name = entry.file_name().into_string().ok()?;
                let path = entry.path();
                if path.extension().map(|e| e.as_bytes() != b"pbd")? {
                    return None;
                }
                let data: toml::Value =
                    toml::from_str(&std::fs::read_to_string(&path).ok()?)
                        .ok()?;

                let environment = data
                    .get("environment")
                    .and_then(|e| e.clone().try_into().ok())
                    .unwrap_or_default();

                let method =
                    data.get("meta")
                        .and_then(|meta| {
                            meta.get("method").and_then(|method| {
                                method.as_str().and_then(|s| match s {
                            "simple-command" => {
                                let benchmark = data["simple-command"]
                                    .clone()
                                    .try_into()
                                    .ok()?;
                                Some(BenchmarkMethod::SimpleCommand(benchmark))
                            }
                            "mercurial-perf-extension" => {
                                let benchmark = data["hg-perf-ext"]
                                    .clone()
                                    .try_into()
                                    .ok()?;
                                Some(BenchmarkMethod::MercurialPerfExtension(
                                    benchmark,
                                ))
                            }
                            _ => None,
                        })
                            })
                        })
                        .unwrap_or(BenchmarkMethod::Unknown);

                Some(Benchmark {
                    path,
                    suite_name: suite_name.to_owned(),
                    environment,
                    name,
                    method,
                    data,
                })
            }))
        };

        let mut bin_env_setups = vec![];

        for setup_file in bin_env_setup_paths {
            let is_python =
                setup_file.extension().map(|e| e == "py").unwrap_or(false);

            let res = if is_python {
                let python_with_poulpe = den_root.join(".venv/bin/python");
                std::process::Command::new(python_with_poulpe)
                    .arg(setup_file.as_os_str())
                    .arg("--describe-parameters")
                    .output()?
            } else {
                std::process::Command::new(setup_file.as_os_str())
                    .arg("--describe-parameters")
                    .output()?
            };

            let output = String::from_utf8_lossy(&res.stdout).to_string();
            let mut description: BinEnvSetup =
                serde_json::from_str(&output).unwrap();
            description.path = setup_file;
            bin_env_setups.push(description)
        }

        Ok(Suite {
            name: suite_name.to_owned(),
            data_env_setups,
            bin_env_setups,
        })
    }
}

pub fn gather_suite(
    den_root: &Path,
    benchmarks: &mut Vec<Benchmark>,
    entry: Result<std::fs::DirEntry, std::io::Error>,
) -> Option<Suite> {
    let entry = entry.ok()?;
    let root = entry.path();
    let stat = std::fs::metadata(&root).ok()?;

    if !stat.is_dir() {
        return None;
    }

    let file_name = entry.file_name();
    // TODO display errors somehow?
    Suite::new(den_root, benchmarks, root, file_name).ok()
}
