//! Miscellaneous utils.

/// You should probably prefer using [`escape_string`]
///
/// Escape the **chars** in `input` for each delimiter in `delimiters` with
/// `escape_seq`, unless the given delimiter is already escaped.
///
/// # Panics
///
/// Panics if the resulting string is not valid UTF8.
///
/// # Examples
///
/// ```
/// use poulpe_core::utils::escape_string_with;
///
/// assert_eq!(escape_string_with("hello", ".", r"\"), "hello");
/// assert_eq!(
///     escape_string_with("hello.test", ".", r"\"),
///     r"hello\.test"
/// );
/// assert_eq!(
///     escape_string_with(r"hello\.test", ".", r"\"),
///     r"hello\.test"
/// );
/// assert_eq!(
///     escape_string_with(r"hello.test", ".", "."),
///     r"hello..test"
/// );
/// assert_eq!(
///     escape_string_with(r"test.", ".", "."),
///     r"test.."
/// );
/// assert_eq!(
///     escape_string_with(r".test.", ".", r"\"),
///     r"\.test\."
/// );
/// assert_eq!(
///     escape_string_with(r"https://test.com&thing=other", ":=", r"\"),
///     r"https\://test.com&thing\=other"
/// );
/// assert_eq!(
///     escape_string_with(
///         r"https://test.com&thing=other",
///         ":=",
///         "DELIMITER"
///     ),
///     r"httpsDELIMITER://test.com&thingDELIMITER=other"
/// );
/// ```
pub fn escape_string_with(
    input: &str,
    delimiters: &str,
    escape_seq: &str,
) -> String {
    let mut to_escape: Vec<char> = input.chars().collect();
    let escape_seq: Vec<char> = escape_seq.chars().collect();
    for delim in delimiters.chars() {
        let mut prev = if to_escape.len() < escape_seq.len() {
            Some(&to_escape[..escape_seq.len()])
        } else {
            None
        };
        let mut inner_escaped = Vec::with_capacity(to_escape.len());
        for (idx, byte) in to_escape.iter().copied().enumerate() {
            if byte == delim && prev != Some(&escape_seq) {
                inner_escaped.extend(&escape_seq);
            }
            inner_escaped.push(byte);
            prev = if idx + escape_seq.len() < to_escape.len() {
                Some(&to_escape[idx..idx + escape_seq.len()])
            } else {
                None
            }
        }
        to_escape = inner_escaped
    }
    to_escape.into_iter().collect()
}

/// Escape the **chars** in `input` for each delimiter in `delimiters` with
/// `\`, unless the given delimiter is already escaped.
///
/// If you want to use a different escape, use [`escape_string_with`]
///
/// # Panics
///
/// Panics if the resulting string is not valid UTF8.
pub fn escape_string(input: &str, delimiters: &str) -> String {
    escape_string_with(input, delimiters, r"\")
}

/// Split a string along `delim`, only if not escaped by `escape`
///
///
/// ['foo$']
/// ```
/// # use poulpe_core::utils::split_unescape;
/// assert_eq!(split_unescape("foo,bar", ',', '\\'), vec!["foo", "bar"]);
/// assert_eq!(split_unescape("foo$,ba$r", ',', '$'), vec!["foo,ba$r"]);
/// assert_eq!(split_unescape("foo$", ',', '$'), vec!["foo$"]);
/// assert_eq!(split_unescape("foo$$,bar", ',', '$'), vec!["foo$", "bar"]);
/// ```
pub fn split_unescape(s: &str, delim: char, escape: char) -> Vec<String> {
    let mut ret = Vec::new();
    let mut current = String::new();
    let mut itr = s.chars();

    while let Some(ch) = itr.next() {
        if ch == escape {
            if let Some(next_char) = itr.next() {
                // Skip the next character if it is escaped
                if next_char == delim {
                    current.push(next_char);
                } else {
                    if next_char != escape {
                        current.push(escape);
                    }
                    current.push(next_char);
                }
            } else {
                current.push(ch);
            }
        } else if ch == delim {
            // Split! (add current to the list and reset it)
            ret.push(current);
            current = String::new();
        } else {
            current.push(ch);
        }
    }

    ret.push(current);
    ret
}
