use std::error::Error;

use egui::{Align, Context, Key, Layout, RichText, TopBottomPanel, Ui};
use egui_commonmark::CommonMarkCache;
use include_dir::{include_dir, Dir};
use poulpe_core::{den::Den, gather_den};

use crate::{
    elements::{benchmark_form::RunBenchmarkForm, menu::render_menu},
    tabs::{
        den_tab::render_den_tab, docs_tab::render_docs_tab,
        run_tab::render_run_tab,
    },
    IS_WEB, WARNING_COLOR,
};

/// Compile-time inclusion of the docs so we can render them inline
pub static DOCS_DIR: Dir<'static> = include_dir!("docs");

#[derive(Clone, Copy, Debug, PartialEq, enum_iterator::Sequence)]
pub enum DashboardMode {
    Docs,
    Den,
    Run,
}

pub struct Dashboard {
    pub den: Option<Den>,
    pub benchmark_form: RunBenchmarkForm,
    pub mode: DashboardMode,
    pub den_path_input: String,
    pub current_doc_page: String,
    /// Time in seconds relative to whatever (used for delta)
    pub last_update: f64,
    /// Cache for the rendered markdown docs
    pub commonmark_cache: CommonMarkCache,
}

impl Dashboard {
    /// Called once before the first frame.
    pub fn new(cc: &eframe::CreationContext<'_>, root: String) -> Self {
        cc.egui_ctx.options_mut(|options| {
            options.max_passes = std::num::NonZeroUsize::new(3).unwrap();
        });

        cc.egui_ctx.style_mut(|style| {
            // Disable text wrapping
            //
            // egui text layouting tries to utilize minimal width possible
            style.wrap_mode = Some(egui::TextWrapMode::Extend);
            style.interaction.tooltip_delay = 0.3;
            // Uncomment this to find out the code responsible for the widget
            // you're hovering. (doesn't work on web, nor with `--release`)
            // style.debug.debug_on_hover = true;
        });

        let mut commonmark_cache = CommonMarkCache::default();
        for file in DOCS_DIR.files() {
            if let Some(p) = file.path().to_str() {
                commonmark_cache.add_link_hook(p)
            }
        }
        let den = gather_den(&root).ok();
        let mode = if den.is_some() {
            DashboardMode::Run
        } else {
            DashboardMode::Docs
        };
        Self {
            den,
            den_path_input: root,
            last_update: 0.0,
            benchmark_form: RunBenchmarkForm::default(),
            commonmark_cache,
            current_doc_page: "poulpe-den.md".to_string(),
            mode,
        }
    }

    pub fn reload_den(&mut self) -> Result<(), Box<dyn Error>> {
        self.den = Some(gather_den(&self.den_path_input)?);
        self.benchmark_form = Default::default();
        Ok(())
    }

    pub fn no_den_message(&mut self, ui: &mut Ui) {
        if !IS_WEB {
            ui.label("Use the top bar to point the dashboard to a den");
        } else {
            ui.colored_label(
                WARNING_COLOR,
                "Remote loading of a den is not available yet",
            );
        }
        if ui.button("Open the docs").clicked() {
            self.mode = DashboardMode::Docs;
        }
    }
}

impl eframe::App for Dashboard {
    /// Called each time the UI needs repainting, which may be many times
    /// per second.
    fn update(&mut self, ctx: &Context, _frame: &mut eframe::Frame) {
        let now = ctx.input(|i| i.time);

        let dt = now - self.last_update;
        let fps = 1.0 / dt;
        let fps_string = format!("{} fps", fps.round());

        handle_zoom(ctx);

        self.last_update = now;

        TopBottomPanel::top("top_panel").show(ctx, |ui| {
            ui.with_layout(Layout::left_to_right(Align::Center), |ui| {
                egui::menu::bar(ui, |ui| render_menu(self, ctx, ui));
            })
        });

        TopBottomPanel::bottom("bottom-panel").show(ctx, |ui| {
            ui.with_layout(Layout::left_to_right(Align::Center), |ui| {
                for mode in enum_iterator::all::<DashboardMode>() {
                    let mut text = RichText::new(format!("{mode:?}"));
                    if mode == self.mode {
                        text = text.strong();
                    }
                    if ui.button(text).clicked() {
                        self.mode = mode;
                    }
                }
                ui.separator();
                ui.with_layout(Layout::right_to_left(Align::Center), |ui| {
                    if ctx.pixels_per_point() != 1.0
                        && ui
                            .button("Reset zoom")
                            .on_hover_text("Reset the global zoom")
                            .clicked()
                    {
                        ctx.set_pixels_per_point(1.0);
                    }

                    egui::warn_if_debug_build(ui);
                    egui::widgets::global_theme_preference_buttons(ui);
                    ui.label(fps_string);
                });
            });
        });

        match self.mode {
            DashboardMode::Docs => render_docs_tab(self, ctx),
            DashboardMode::Den => render_den_tab(self, ctx),
            DashboardMode::Run => render_run_tab(self, ctx),
        }
    }
}

fn handle_zoom(ctx: &Context) {
    let (scroll_zoom, reset) = ctx.input(|i| {
        (i.zoom_delta(), i.modifiers.ctrl && i.key_pressed(Key::Num0))
    });
    if scroll_zoom != 1.0 {
        let sign = if scroll_zoom < 1.0 {
            -1.0
        } else if scroll_zoom == 0.0 {
            0.0
        } else {
            1.0
        };
        ctx.set_pixels_per_point(ctx.pixels_per_point() + 0.005 * sign);
    }
    if reset {
        ctx.set_pixels_per_point(1.0);
    }
}
