use std::collections::HashMap;

use egui::{
    Align, Button, ComboBox, FontId, Key, Layout, Response, RichText,
    ScrollArea, TextEdit, Ui, WidgetText,
};
use egui_taffy::{taffy::prelude::*, Tui, TuiBuilderLogic};
use poulpe_core::{
    benchmark::{Benchmark, BenchmarkMethod},
    binary_environment::Setup as BinEnvSetup,
    data_environment::DataEnvironment,
    den::Den,
    scheduling::build_scheduler_run_instruction,
};

use crate::{ERROR_COLOR, WARNING_COLOR};

type FormField = String;
type FormFields = HashMap<FormField, String>;

pub struct RunBenchmarkForm {
    /// Index (within our [`Den`] of the currently selected [`Benchmark`]
    selected_benchmark_idx: usize,
    /// Current form data for each [`Benchmark`], filled lazily
    form_data_for_benchmark: HashMap<String, BenchmarkFormData>,
    /// The schedule file we're adding to
    schedule: String,
}

impl Default for RunBenchmarkForm {
    fn default() -> Self {
        Self {
            selected_benchmark_idx: Default::default(),
            form_data_for_benchmark: Default::default(),
            schedule: String::from(
                "# Use the form to add scheduler instructions.\n\
                 # \n\
                 # These instructions should go into a schedule file \
                 (by default `$DEN_DIR/current.poulpe-schedule`) \
                 and be run with `poulpe den-execute-tasks`.\n",
            ),
        }
    }
}

impl RunBenchmarkForm {
    pub fn render_schedule(&mut self, ui: &mut Ui) {
        ui.with_layout(Layout::top_down(Align::Center), |ui| {
            ui.label("Schedule file");
            ScrollArea::vertical().show(ui, |ui| {
                ui.add_sized(
                    ui.available_size(),
                    TextEdit::multiline(&mut self.schedule),
                );
            });
        });
    }

    pub fn render_form(&mut self, tui: &mut Tui, den: &Den) {
        let benchmarks = &den.benchmarks;
        if benchmarks.is_empty() {
            tui.colored_label(WARNING_COLOR, "No benchmarks available");
            return;
        }
        if den.data_envs.is_empty() {
            tui.colored_label(WARNING_COLOR, "No data-envs available");
            return;
        }
        let selected = &benchmarks[self.selected_benchmark_idx];
        let mut form_state = FormState::default();

        let heading = tui.heading("Benchmark to run");
        Self::add_combobox(
            tui,
            &mut form_state,
            "Benchmark to run",
            selected.qualified_name(),
            |ui, form_state| {
                for (idx, benchmark) in benchmarks.iter().enumerate() {
                    let name = benchmark.qualified_name();
                    form_state.update(&ui.selectable_value(
                        &mut self.selected_benchmark_idx,
                        idx,
                        name,
                    ));
                }
            },
        );

        let Some(suite) = den.find_suite_by_name(&selected.suite_name) else {
            tui.colored_label(
                WARNING_COLOR,
                format!("No setup found for suite '{}'", selected.suite_name),
            );
            return;
        };

        if suite.bin_env_setups.is_empty() {
            tui.colored_label(
                WARNING_COLOR,
                format!(
                    "No bin-env setup found for suite '{}'",
                    selected.suite_name
                ),
            );
            return;
        }
        let setup = &suite.bin_env_setups[0];
        // Used to ensure form values and egui IDs are namespaced by both the
        // benchmark and the setup used.
        let form_namespace_key =
            format!("{} {}", selected.path.display(), setup.path.display());
        heading.on_hover_text(format!(
            "Using bin-env setup at {}",
            &form_namespace_key,
        ));
        let fields = self
            .form_data_for_benchmark
            .entry(form_namespace_key.clone())
            .or_insert_with(|| {
                let benchmark_fields = Default::default();
                let compatible_data_envs =
                    get_compatible_data_envs(den, selected, &benchmark_fields);
                let selected_data_env = if compatible_data_envs.is_empty() {
                    None
                } else {
                    Some(compatible_data_envs[0].clone())
                };
                BenchmarkFormData {
                    selected_data_env,
                    compatible_data_envs,
                    bin_env_fields: Default::default(),
                    benchmark_fields,
                    errors: Default::default(),
                }
            });

        tui.separator();
        tui.heading("Binary environment parameters");
        render_benchmark_variants(
            tui,
            selected,
            &mut form_state,
            &mut fields.benchmark_fields,
            &form_namespace_key,
        );

        tui.separator();
        tui.heading("Benchmark variants");

        let bin_env_fields = &mut fields.bin_env_fields;

        for (name, info) in &setup.parameters {
            Self::with_field_style(tui, |tui| {
                let label = if info.default.is_some() {
                    tui.label(name)
                } else {
                    tui.colored_label(
                        WARNING_COLOR,
                        format!("{name} (required)"),
                    )
                };
                if let Some(errors) = fields.errors.get(name) {
                    for error in errors {
                        tui.colored_label(ERROR_COLOR, error.to_string());
                    }
                }
                if let Some(doc) = &info.doc {
                    label.on_hover_text(doc);
                }
                let salt = format!("{} {}", form_namespace_key, name);
                if let Some(choices) = &info.choices {
                    let selected =
                        bin_env_fields.entry(name.clone()).or_insert_with(
                            || (info.default).clone().unwrap_or_default(),
                        );
                    Self::add_combobox(
                        tui,
                        &mut form_state,
                        salt,
                        selected.clone(),
                        |ui, form_state| {
                            for (choice_name, doc) in choices.iter() {
                                let response = ui.selectable_value(
                                    selected,
                                    choice_name.clone(),
                                    choice_name,
                                );
                                form_state.update(&response);
                                response.on_hover_text(doc);
                            }
                        },
                    );
                } else {
                    let text = TextEdit::singleline(
                        bin_env_fields.entry(name.clone()).or_insert_with(
                            || info.default.clone().unwrap_or("".to_string()),
                        ),
                    )
                    .desired_width(tui.egui_ui().available_width())
                    .id_salt(salt)
                    .return_key(None)
                    .clip_text(true);
                    let text = tui.ui_add(text);

                    form_state.update(&text);

                    if let Some(doc) = &info.doc {
                        text.on_hover_text(doc);
                    }
                }
            })
        }

        tui.separator();
        tui.heading("Data environment");
        let compatible_data_envs = &fields.compatible_data_envs;
        let has_compatible_data_envs = !compatible_data_envs.is_empty();
        if !has_compatible_data_envs {
            tui.colored_label(
                WARNING_COLOR,
                "No compatible data environment \
                 found for this benchmark configuration",
            );
        } else {
            let selected_data_env = fields
                .selected_data_env
                .as_mut()
                .expect("there should be at least one valid data-env");

            Self::add_combobox(
                tui,
                &mut form_state,
                "Target data-env",
                selected_data_env.name.to_owned(),
                |ui, form_state| {
                    for data_env in compatible_data_envs {
                        form_state.update(&ui.selectable_value(
                            &mut selected_data_env.name,
                            data_env.name.to_owned(),
                            data_env.name.to_owned(),
                        ));
                    }
                },
            );
        }
        let setup = &suite.bin_env_setups[0];

        let add_benchmark_button = tui
            .style(Style {
                align_self: Some(AlignItems::Center),
                padding: length(10.0),
                ..Default::default()
            })
            .add(|tui| {
                tui.wrap_mode(egui::TextWrapMode::Extend).add(|tui| {
                    let add_benchmark_button =
                        Button::new(RichText::new("Add benchmark").font(
                            FontId::new(15.0, egui::FontFamily::Proportional),
                        ));
                    if !has_compatible_data_envs {
                        tui.disabled().ui_add(add_benchmark_button)
                    } else {
                        tui.ui_add(add_benchmark_button)
                    }
                })
            });
        let mut clicked = add_benchmark_button.clicked();

        if form_state.any_changed {
            fields.compatible_data_envs = get_compatible_data_envs(
                den,
                selected,
                &fields.benchmark_fields,
            );
        }

        clicked |= form_state.any_focused
            && tui
                .egui_ui_mut()
                .input(|i| i.modifiers.ctrl && i.key_pressed(Key::Enter));

        if clicked {
            if let Some(schedule_line) = fields.submit(setup, selected) {
                self.schedule.push_str(&format!(
                    "{}{}",
                    if self.schedule.is_empty() || self.schedule.ends_with('\n')
                    {
                        ""
                    } else {
                        "\n"
                    },
                    schedule_line,
                ));
            }
        }

        if !fields.errors.is_empty() {
            tui.colored_label(
                ERROR_COLOR,
                format!("Found {} errors", fields.errors.len()),
            );
        }
    }

    fn add_combobox<R>(
        tui: &mut Tui,
        form_state: &mut FormState,
        id: impl std::hash::Hash,
        selected_text: impl Into<WidgetText>,
        contents: impl FnOnce(&mut Ui, &mut FormState) -> R,
    ) {
        let id = tui
            .ui_add_manual(
                |ui| {
                    ComboBox::from_id_salt(id)
                        .selected_text(selected_text)
                        .width(ui.available_width())
                        .show_ui(ui, |ui| contents(ui, form_state))
                        .response
                },
                |value, _| value,
            )
            .id;
        form_state.update_focused(ComboBox::is_open(tui.egui_ui().ctx(), id));
    }

    fn with_field_style(tui: &mut Tui, doc: impl FnMut(&mut Tui)) {
        let field_style = Style {
            flex_direction: FlexDirection::Column,
            gap: length(5.0),
            ..Default::default()
        };
        tui.style(field_style).add(doc);
    }
}

/// Current form data for a given benchmark and bin-env setup
struct BenchmarkFormData {
    /// The data env we've selected, may be `None` if no data env satisfy the
    /// default benchmark parameters
    selected_data_env: Option<DataEnvironment>,
    /// The current set of [`DataEnvironment`] compatible with the selected
    /// benchmark if parametrized with [`Self::benchmark_fields`]
    compatible_data_envs: Vec<DataEnvironment>,
    /// The current values for each parameter of the benchmark
    benchmark_fields: FormFields,
    /// The current values for each parameter that make up the target bin-env
    bin_env_fields: FormFields,
    /// The set of [`ValidationError`] for each field, if any
    errors: HashMap<FormField, Vec<ValidationError>>,
}

impl BenchmarkFormData {
    fn submit(
        &mut self,
        setup: &BinEnvSetup,
        benchmark: &Benchmark,
    ) -> Option<String> {
        self.errors = self.validate(setup);
        if self.errors.is_empty() {
            Some(self.get_schedule_line(setup, benchmark))
        } else {
            None
        }
    }

    fn get_schedule_line(
        &mut self,
        setup: &BinEnvSetup,
        benchmark: &Benchmark,
    ) -> String {
        let bin_env_parameters = self.bin_env_fields.iter().fold(
            vec![],
            |mut acc, (field, value)| {
                let default = &setup.parameters[field].default;
                if !value.is_empty() && Some(value) != default.as_ref() {
                    acc.push((field.as_str(), value.as_str()));
                }
                acc
            },
        );
        let bench_parameters = self.benchmark_fields.iter().fold(
            vec![],
            |mut acc, (field, value)| {
                let default = match &benchmark.method {
                    BenchmarkMethod::SimpleCommand(bench) => bench
                        .variants
                        .as_ref()
                        .and_then(|v| v.default_key(field)),
                    BenchmarkMethod::MercurialPerfExtension(bench) => bench
                        .variants
                        .as_ref()
                        .and_then(|v| v.default_key(field)),
                    BenchmarkMethod::Unknown => None,
                };
                if !value.is_empty() && Some(value.as_str()) != default {
                    acc.push((field.as_str(), value.as_str()));
                }
                acc
            },
        );
        let data_env_path = &self
            .selected_data_env
            .as_ref()
            .expect("there should be a selected data env")
            .root;

        build_scheduler_run_instruction(
            benchmark,
            bin_env_parameters,
            bench_parameters,
            data_env_path,
        )
    }

    fn validate(
        &self,
        setup: &BinEnvSetup,
    ) -> HashMap<FormField, Vec<ValidationError>> {
        let mut errors = HashMap::new();

        for (field, value) in self.bin_env_fields.iter() {
            let field_info = &setup.parameters[field];
            if field_info.default.is_none() && value.is_empty() {
                let field_errors: &mut Vec<_> =
                    errors.entry(field.to_owned()).or_default();
                field_errors.push(ValidationError::RequiredField);
            }
        }
        errors
    }
}

/// Collects the relevant state of the form being displayed in this frame.
/// Useful for user interaction and performance.
#[derive(Debug, Default)]
struct FormState {
    /// True if any field within the form is currently focused/opened
    any_focused: bool,
    // TODO make it work for ComboBox as well?
    /// True if any field within the form has lost focus in the current frame
    any_lost_focus: bool,
    /// True if any field within the form has changed in the current frame.
    /// Change is a text value changing, a combobox value changing, etc., but
    /// not a button being pressed.
    any_changed: bool,
}

impl FormState {
    fn update(&mut self, response: &Response) {
        if response.has_focus() {
            self.any_focused = true
        }
        if response.lost_focus() {
            self.any_lost_focus = true
        }
        if response.changed() {
            self.any_changed = true
        }
    }

    fn update_focused(&mut self, focused: bool) {
        if focused {
            self.any_focused = true;
        }
    }
}

fn get_compatible_data_envs(
    den: &Den,
    selected: &Benchmark,
    parameters: &FormFields,
) -> Vec<DataEnvironment> {
    den.compatible_data_envs(selected, parameters)
        .into_iter()
        .map(|v| (*v).clone())
        .collect()
}

fn render_benchmark_variants(
    tui: &mut Tui,
    benchmark: &Benchmark,
    form_state: &mut FormState,
    benchmark_fields: &mut FormFields,
    form_namespace_key: &str,
) {
    match &benchmark.method {
        BenchmarkMethod::SimpleCommand(simple_command_benchmark) => {
            if let Some(variants) = &simple_command_benchmark.variants {
                for (name, cases) in &variants.dimensions {
                    let Some(default_key) = variants.default_key(name) else {
                        continue;
                    };

                    RunBenchmarkForm::with_field_style(tui, |tui| {
                        let selected = benchmark_fields
                            .entry(name.clone())
                            .or_insert_with(|| default_key.to_owned());
                        tui.label(name);
                        if cases.len() <= 1 {
                            tui.label(default_key);
                        } else {
                            let salt = format!(
                                "variant {} {} {}",
                                form_namespace_key, benchmark.name, name
                            );
                            RunBenchmarkForm::add_combobox(
                                tui,
                                form_state,
                                salt,
                                selected.clone(),
                                |ui, form_state| {
                                    for choice_name in cases.keys() {
                                        form_state.update(
                                            &ui.selectable_value(
                                                selected,
                                                choice_name.clone(),
                                                choice_name,
                                            ),
                                        );
                                    }
                                },
                            );
                        }
                    });
                }
            }
        }
        BenchmarkMethod::MercurialPerfExtension(mercurial_perf_benchmark) => {
            // TODO is this duplicated from simple command for good reason?
            if let Some(variants) = &mercurial_perf_benchmark.variants {
                for (name, cases) in &variants.dimensions {
                    let Some(default_key) = variants.default_key(name) else {
                        continue;
                    };

                    let selected = benchmark_fields
                        .entry(name.clone())
                        .or_insert_with(|| default_key.to_owned());
                    tui.label(name);
                    if cases.len() <= 1 {
                        tui.label(default_key);
                    } else {
                        let salt = format!(
                            "variant {} {} {}",
                            form_namespace_key, benchmark.name, name
                        );
                        RunBenchmarkForm::add_combobox(
                            tui,
                            form_state,
                            salt,
                            selected.clone(),
                            |ui, form_state| {
                                for choice_name in cases.keys() {
                                    form_state.update(&ui.selectable_value(
                                        selected,
                                        choice_name.clone(),
                                        choice_name,
                                    ));
                                }
                            },
                        );
                    }
                }
            }
        }
        BenchmarkMethod::Unknown => {}
    }
}

pub enum ValidationError {
    /// The current field is required and the user did not provide a value
    RequiredField,
}

impl std::fmt::Display for ValidationError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            ValidationError::RequiredField => f.write_str("field is required"),
        }
    }
}
