use egui::{CollapsingHeader, RichText, ScrollArea, Ui};
use poulpe_core::{
    benchmark::{Benchmark, BenchmarkMethod},
    den::Den,
    runner::{
        mercurial_perf::MercurialPerfCommandVariants,
        simple_command::SimpleCommandVariants,
    },
};

use crate::WARNING_COLOR;

pub fn render_den_drawer(den: Option<&Den>, ui: &mut Ui) {
    ScrollArea::vertical().show(ui, |ui| {
        let Some(den) = den else {
            ui.colored_label(WARNING_COLOR, "No den found");
            return;
        };

        CollapsingHeader::new("Bin envs").show(ui, |ui| {
            for bin_env in &den.bin_envs {
                let name = bin_env.name();
                let ready = if bin_env.is_ready {
                    "READY"
                } else {
                    "NOT READY"
                };
                ui.label(format!("{}: {}", name, ready,));
            }
        });

        CollapsingHeader::new("Data envs").show(ui, |ui| {
            for data_env in &den.data_envs {
                ui.label(&data_env.name);
            }
        });

        CollapsingHeader::new("Benchmarks").show(ui, |ui| {
            for benchmark in &den.benchmarks {
                render_benchmark(ui, benchmark);
            }
        });
    });
}

fn render_benchmark(ui: &mut Ui, benchmark: &Benchmark) {
    let qualified_name = benchmark.qualified_name();
    match &benchmark.method {
        BenchmarkMethod::SimpleCommand(simple_command_benchmark) => {
            if let Some(variants) = &simple_command_benchmark.variants {
                CollapsingHeader::new(&qualified_name).show(ui, |ui| {
                    render_simple_command_variants(
                        ui,
                        variants,
                        &qualified_name,
                    )
                });
            } else {
                ui.colored_label(
                    ui.visuals().widgets.inactive.text_color(),
                    qualified_name,
                );
            }
        }
        BenchmarkMethod::MercurialPerfExtension(hg_perf_benchmark) => {
            if let Some(variants) = &hg_perf_benchmark.variants {
                CollapsingHeader::new(&qualified_name).show(ui, |ui| {
                    render_mercurial_perf_extension_variants(
                        ui,
                        variants,
                        &qualified_name,
                    )
                });
            } else {
                ui.colored_label(
                    ui.visuals().widgets.inactive.text_color(),
                    qualified_name,
                );
            }
        }
        BenchmarkMethod::Unknown => {
            ui.colored_label(
                WARNING_COLOR,
                format!("{} (unknown type)", qualified_name),
            );
        }
    }
}

fn render_simple_command_variants(
    ui: &mut Ui,
    variants: &SimpleCommandVariants,
    benchmark_qualified_name: &str,
) {
    for (name, info) in &variants.dimensions {
        let id = format!("{}{}", benchmark_qualified_name, name);
        CollapsingHeader::new(id).show(ui, |ui| {
            for (name, case) in info {
                if case.default {
                    ui.label(
                        RichText::new(format!("{name} (default)")).strong(),
                    );
                } else {
                    ui.label(name);
                }
            }
        });
    }
}

fn render_mercurial_perf_extension_variants(
    ui: &mut Ui,
    variants: &MercurialPerfCommandVariants,
    benchmark_qualified_name: &str,
) {
    for (name, info) in &variants.dimensions {
        let id = format!("{}{}", benchmark_qualified_name, name);
        CollapsingHeader::new(id).show(ui, |ui| {
            for (name, case) in info {
                if case.default {
                    ui.label(
                        RichText::new(format!("{name} (default)")).strong(),
                    );
                } else {
                    ui.label(name);
                }
            }
        });
    }
}
