extends "benchmark.pkl"

method = "simple-command"

/// The command to be run.
///
/// It can we extended with variants (see next section)
command: String

/// A command (or series of commands) to run before the entire benchmark
/// to gather runtime information (exact version of a remote service, etc.)
/// to save in the results.
/// The script will be run in the same shell and environment than the benchmark.
/// It should write its output to `"$POULPE_TMP_DIR"/gathered_info.json`
/// as valid JSON data.
gather_info: String?

/// A command (or series of commands) to run before each command run.
prepare_run: String?

/// A command (or series of commands) to run after each command run.
cleanup_run: String?

/// Variants allow for slight variation of a benchmark. For example, they allow to change the
/// input data, of to turn some feature on and off.
///
/// Variants are organised within independant "dimensions".
///
/// Each dimension has a set of cases, each value can modify the benchmark. Each
/// dimension must have a default value.
variants: Mapping<String, SimpleCommandVariant>

/// Ignore command failure when running benchmark
accept_failure: Boolean = false

///////////////////////////////////////////////////////////////////////////////
/// rendering as toml /////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
as_v0 = (super.as_v0) {
    ["simple-command"] = new Dynamic {
        when (gather_info != null) {
            ["gather-info"] = gather_info
        }
        ["command"] = command
        when (prepare_run != null) {
            ["prepare-run"] = prepare_run
        }
        when (cleanup_run != null) {
            ["cleanup-run"] = cleanup_run
        }
        when (accept_failure) {
            ["accept-failure"] = accept_failure
        }
        when (module.cwd != null) {
            ["cwd"] = module.cwd
        }
        when (module.constraints != null) {
            ["constraints"] = module.constraints.as_v0
        }
        when (!variants.isEmpty) {
            ["variants"] = new Mapping {
                ["dimensions"] = new Mapping {
                    for (d, v in variants) {
                        [d] = v.as_v0
                    }
                }
            }
        }
    }
}

///////////////////////////////////////////////////////////////////////////////
/// base classes and utilities ////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
fixed hidden blah = null

/// Variants are defined under keys following this patterns :
///
///   simple-command.variants.dimensions.<dimension-name>.<variant-key>
///
/// The `dimension-name` and `variant-key` are used as identifier in the
/// benchmark / result and for selecting variant at run time.
class SimpleCommandVariant {
    /// The variant that should be used by default
    default_key: String((s) -> cases.containsKey(s))
    cases: Mapping<String, SimpleCommandVariantCase>
    isEmpty = cases.isEmpty

    /// render as "v0" format
    hidden fixed as_v0: Dynamic = new {
        // XXX assert there is one default key
        for (key, value in cases) {
            [key] = (value) {
                default = key == default_key
            }.as_v0
        }
    }
}

class SimpleCommandVariantCase {
    /// A section that allow to set environment variable to be used when
    /// running the command. This is useful to replace some of the command
    /// argument with. Each sub-key is a an environment variable to replace.
    ///
    /// Value supports the `DATA-VARS:` notation if necessary
    environment: Environment

    /// Exactly one variant must have it set to `true` per dimension.
    ///
    /// When set and no variant have been explicitly selected for this dimension, this
    /// variant will be used.
    default: Boolean = false

    /// A string to append to the command when this variable is selected.
    extend_command: String?

    /// overwrite the global `simple-command.cwd` variable. Behave the same.
    cwd: String?

    /// return code that should be considered a success
    acceptable_return_codes: Listing<Int>?

    /// additional constraint on when this variant can be used
    constraints: Constraints?

    /// render as "v0" format
    hidden fixed as_v0: Dynamic = new {
        when (!environment.isEmpty) {
            ["environment"] = environment
        }
        when (default) {
            ["default"] = default
        }
        when (extend_command != null) {
            ["extend-command"] = extend_command
        }
        when (cwd != null) {
            ["cwd"] = cwd
        }
        when (acceptable_return_codes != null) {
            ["acceptable-return-codes"] = acceptable_return_codes
        }
        when (constraints != null) {
            ["constraints"] = constraints.as_v0
        }
    }
}
