import "perf.pkl" as p
import "simple-command.pkl" as sc
// Gather common variants that are likely to exists in multiple benchmark

perf: Mapping<String, p.PerfVariant> = new {
    ["revlog"] {
        default_key = "manifestlog"
        cases {
            ["changelog"] {
                args {
                    "--changelog"
                }
            }
            ["manifestlog"] {
                args {
                    "--manifest"
                }
            }
        }
    }
    ["stream-version"] {
        default_key = "latest"
        cases {
            for (v in List("latest", "v1", "v2", "v3")) {
                [v]{
                    args{
                        "--stream-version"
                        v + if (v == "v3") "-exp" else ""
                    }
                }
            }
        }
    }
    ["issue6528"] {
        default_key = "default"
        cases {
            ["default"] {}
            ["fix-incoming"] {
                args {
                    "--config"
                    "storage.revlog.issue6528.fix-incoming=yes"
                }
                constraints {
                    compatible_revision_range = "2813d406b036::"
                }
            }
            ["disabled"] = (this["fix-incoming"]){
                args{
                    [1] = (super[1]).replaceLast("=yes", "=no")
                }
            }
        }
    }
    ["bundled-revs"] {
        // leaving default_key choice to the user
        cases {
            for (r in List(
                1,
                10,
                100,
                1000,
                10000,
                100000
            )){
                ["any-\(r)-extra-rev"] {
                    args {
                        "DATA-VARS:mercurial.bundles.any-\(r)-revs"
                    }
                    // some 2**N behavior prevent reasonable pulling time for set of revision
                    // introduced in 23950e39281f, fixed in c9ceb4f60256
                    when (r > 10) {
                        constraints {
                            compatible_revision_range = "not (23950e39281f:: and not c9ceb4f60256::)"
                        }
                    }
                }
            }
        }
    }
    ["resource-usage"] {
        default_key = "default"
        cases {
            ["default"] {}
            for (u in List(
                "high",
                "medium",
                "low"
            )) {
                [u] {
                    args {
                        "--config"
                        "usage.resources=\(u)"
                    }
                }
            }
        }
    }
    for (resource_type in List("cpu", "memory", "disk")) {
        ["resources-\(resource_type)"] {
            default_key = "default"
            cases {
                ["default"] {}
                for (u in List("low", "medium", "high")) {
                    [u] {
                        args {
                            "--config"
                            "usage.resources.\(resource_type)=\(u)"
                        }
                    }
                }
            }
        }
    }
}

simple_command: Mapping<String, sc.SimpleCommandVariant>  = new {
    ["verbosity"] {
        default_key = "standard"
        cases {
            ["quiet"] {extend_command = " --quiet"}
            ["standard"] {extend_command = ""}
            ["verbose"] {extend_command = " --verbose"}
            ["debug"] {extend_command = " --debug"}
        }
    }
    ["revlog"] {
        default_key = "manifestlog"
        cases {
            ["changelog"] {
                extend_command = " --changelog"
            }
            ["manifestlog"] {
                extend_command = " --manifest"
            }
        }
    }
    ["rev-reading-target"] {
        default_key = "default"
        cases {
            ["working-copy"]{
                environment {
                    ["REV"] = "-r wdir()"
                }
            }
            ["tip"]{
                environment {
                    ["REV"] = "-r tip"
                }
            }
            ["default"]{
                environment {
                    ["REV"] = ""
                }
            }
        }
    }
    ["rev-reading-output-format"] {
        default_key = "plain"
        cases {
            ["plain"] {
                environment {
                    ["TEMPLATE"] = ""
                }
            }
            ["json"] {
                environment {
                    ["TEMPLATE"] = "--template json"
                }
            }
        }
    }

    // variante providing files patterns to access in a single revision
    //
    // used for example for `hg cat`, or `hg files`
    ["rev-reading-files"] {
        default_key = "all-root"
        cases {
            ["all-root"] {
                environment {
                    ["FILES"] = "."
                }
            }
            ["all-glob"] {
                environment {
                    ["FILES"] = "set:**"
                }
            }

            ["all-list-path-sorted"] {
                environment {
                    ["FILES"] = "set:listfile:../all_sorted.path"
                }
            }
            ["all-list-path-shuffled"] {
                environment {
                    ["FILES"] = "set:listfile:../all_suffled.path"
                }
            }
            ["all-list-filepath-sorted"] {
                environment {
                    ["FILES"] = "set:listfile:../all_sorted.filepath"
                }
            }
            ["all-list-filepath-shuffled"] {
                environment {
                    ["FILES"] = "set:listfile:../all_suffled.filepath"
                }
            }

            ["all-list-pattern-sorted"] {
                environment {
                    ["FILES"] = "set:listfile:../all_sorted.pattern"
                }
            }

            ["all-list-pattern-shuffled"] {
                environment {
                    ["FILES"] = "set:listfile:../all_suffled.pattern"
                }
            }
        }
    }
    // single file that has an interresting file history
    //
    // to be used by thing like `hg annotate` or `hg log`
    ["file-reading-revs"] {
        default_key = "all-root"
        cases {
            for (suffix in List("a", "b", "c", "d", "e")) {
                ["file-with-long-history.\(suffix)"] {
                    environment {
                        ["FILES"] = "DATA-VARS:mercurial.file-with-long-history.\(suffix)"
                    }
                }
            }
        }
    }
    // repo state should probably be split into multiple dimensions.
    //
    // However, that would be harder to do with the currently limited variant
    // support.
    ["repo-state"] {
        // explicitly lacking a default_key, user need to define one
        cases {
            for (state in List(
                "modified",
                "clean",
                "large-ignored",
                // Worst case scenario
                "large-all",
                "ignored",
                "ignored-unknowns",
                "all-and-large-ignored",
                "large-modified",
                "large-modified-and-copied",
                "modified-and-copied"
            )) {
                [state] {
                    cwd = "DATA-VARS:mercurial.working-copy.\(state)"
                }
            }
        }
    }
    ["resource-usage"] {
        default_key = "default"
        cases {
            for (u in List(
                "default",
                "high",
                "medium",
                "low"
            )) {
                [u] {extend_command = " --config usage.resources=\(u)"}
            }
        }
    }
    ["exchange-explicit-rev"] {
        default_key = "none"
        cases {
            ["none"] {}
            ["tip"] {
                extend_command = " --rev tip"
            }
            ["all-out-heads"] {
                extend_command = " `cat OUT_NODES.txt`"
            }
        }
    }
    ["bundled-revs"] {
        //let default_key the responsability of users
        cases {
            for (k, v in perf["bundled-revs"].cases) {
                [k] {
                    constraints = v.constraints
                    environment {
                        ["BUNDLE_PATH"] = v.args[0]
                    }

                }
            }
        }
    }
    ["validate-incoming"] {
        default_key = "default"
        cases {
            ["default"] {}
            for (v in List("no", "yes")) {
                [v] {extend_command=" --config server.validate=\(v)"}
            }
        }
    }
    ["issue6528"] {
        default_key = perf["issue6528"].default_key
        cases {
            for (k, v in perf["issue6528"].cases) {
                [k] {
                    environment {
                        ["ISSUE6528_ARG"] = v.args.toList().join(" ")
                    }
                    constraints = v.constraints
                }
            }
        }
    }
    ["issue6528-config"] {
        default_key = "default"
        cases {
            ["default"] {
                environment {
                    ["ISSUE6528_CONFIG"]=""
                }
            }
            ["fix-incoming"] {
                environment {
                    ["ISSUE6528_CONFIG"]="""
                    [storage]
                    revlog.issue6528.fix-incoming=yes
                    """
                }
                constraints {
                    compatible_revision_range = "2813d406b036::"
                }
            }
            ["disabled"] = (this["fix-incoming"]){
                environment {
                    ["ISSUE6528_CONFIG"] = """
                        [storage]
                        revlog.issue6528.fix-incoming=no
                        """
                }
                constraints {
                    compatible_revision_range = "2813d406b036::"
                }
            }
        }
    }
    ["pulled-delta-reuse-policy"] {
        default_key = "default"
        cases {
            ["default"] {
                environment {
                    ["PULLED_DELTA_REUSE_CFG"] = ""
                }
            }
            ["forced"] {
                environment {
                    ["PULLED_DELTA_REUSE_CFG"] = "--config 'paths.*:pulled-delta-reuse-policy=forced'"
                }
                constraints {
                    compatible_revision_range = "f35cf52acabd::"
                }
            }
        }

    }
    ["reuse-external-delta-parent-config"] {
        default_key = "default"
        cases {
            ["default"] {
                environment {
                    ["DELTA_PARENT_CONFIG"] = ""
                }
            }
            ["yes"] {
                environment {
                    ["DELTA_PARENT_CONFIG"] = """
                        [storage]
                        revlog.reuse-external-delta-parent=yes
                        """
                }
                constraints {
                    compatible_revision_range = "f6eff9e4de80::"
                }
            }
            ["no"] {
                environment {
                    ["DELTA_PARENT_CONFIG"] = """
                        [storage]
                        revlog.reuse-external-delta-parent=no
                        """
                }
                constraints {
                    compatible_revision_range = "f6eff9e4de80::"
                }
            }
        }
    }
    ["protocol"] {
        default_key = "local"
        cases {
            ["local"] {
                environment {
                    ["PROTO"] = "local"
                }
            }
            ["ssh"] {
                environment {
                    ["PROTO"] = "ssh"
                }
                extend_command = """
                     --config ui.remotecmd="$POULPE_SSH_REMOTE_HG"
                    """
            }
        }
    }
}

