#!/bin/bash
#
# WARNING: Do not run this script directly, use `poulpe bin-env-util setup-one`
#
# environment variable that MUST be set:
#
#   MERCURIAL_VERSION: an identifier of the Mercurial version (or changeset) we want to install,
#                      This will be used to run `hg update`.
#
# environment variable that MAY be set:
#
#   PYTHON_VERSION     a path to the Python to use for the bin env
#   MERCURIAL_REPO_URL a path to a mercurial clone.
#   MERCURIAL_FLAVOR   a type of installation, available value:
#                          - default: Mercurial as you would get it by default
#                                     for that platform and version.
#                          - pure:    Pure Python implementation of Mercurial
#                          - no-rust: Mercurial with CExt compiled extension,
#                                     but no rust ones.
#                          - rust:    Mercurila with all available compiled extensions,
#                                     including the Rust one
#                          - rhg:     The pure Rust fastpath, (with fallback when needed)
#   MERCURIAL_PY_RE2 the name of a python binding to the re2 library to install.
#                      Known good value are :
#                          - google-re2
#
# the script work use some hardcoded temporary directory :
#   - /tmp/poulpe-tmp/hg
#   - /tmp/poulpe-clone-cache/hg

set -Eeuo pipefail

if [ "$#" -gt 0 ]; then
if [ "$1" = "--describe-parameters" ]; then
cat << EOF
{
    "parameters": {
        "MERCURIAL_VERSION": {
            "doc": "an identifier of the Mercurial version we want to install",
            "type": "<revspec>",
            "arg-mode": "ENV"
        },
        "PYTHON_VERSION": {
            "default": "python3",
            "doc": "a path to the Python to use for the bin env",
            "type": "<file-path>",
            "arg-mode": "ENV"
        },
        "MERCURIAL_REPO_URL": {
            "default": "https://foss.heptapod.net/mercurial/mercurial-devel/",
            "doc": "a path to a mercurial clone",
            "type": "<url>",
            "arg-mode": "ENV"
        },
        "MERCURIAL_FLAVOR": {
            "default": "default",
            "doc": "the type of installation",
            "type": "choice",
            "choices": {
                "default": "Mercurial as you would get it by default for that platform and version",
                "pure": "Pure Python implementation of Mercurial",
                "no-rust": "Mercurial with Cext compiled extension but no rust ones",
                "rust": "Mercurial with all available compiled extensions, including the Rust one",
                "rhg": "the pure Rust fastpath, (with fallback when needed)"
            },
            "arg-mode": "ENV"
        },
        "MERCURIAL_PY_RE2": {
            "default": "none",
            "doc": "Name of an optional python binding to the re2 library to install",
            "type": "choice",
            "choices": {
                "google-re2": "Use the google-re2 module for binding",
                "none": "Don't use an re2 binding"
            },
            "arg-mode": "ENV"
        },
        "EVOLVE_VERSION": {
            "doc": "an optional identifier of the evolve version we want to install",
            "default": "",
            "type": "<revspec>",
            "arg-mode": "ENV"
        },
        "EVOLVE_REPO_URL": {
            "default": "https://foss.heptapod.net/mercurial/evolve/",
            "doc": "a path to a evolve clone",
            "type": "<url>",
            "arg-mode": "ENV"
        }
    }
}
EOF
exit 0
fi
fi


export HGRCPATH=
export HGPLAIN=1
unset HGWITHRUSTEXT
export HGWITHRUSTEXT
unset HGMODULEPOLICY
export HGMODULEPOLICY

repo_url="${MERCURIAL_REPO_URL:-https://foss.heptapod.net/mercurial/mercurial-devel}"
evolve_repo_url="${EVOLVE_REPO_URL:-https://foss.heptapod.net/mercurial/evolve}"
flavor="${MERCURIAL_FLAVOR:-default}"
pyre2="${MERCURIAL_PY_RE2:-default}"

evolve_version="${EVOLVE_VERSION:-}"

[ -n "${MERCURIAL_VERSION}" ]

python_override="${PYTHON_VERSION:-python3}"

echo "POULPE: using Python $python_override"
echo "POULPE: Python version `$python_override --version`"
echo "POULPE: Python executable path `which $python_override`"
echo "POULPE: Python executable real path `$python_override -c 'import sys; print(sys.executable)'`"

mkdir --parents poulpe-tmp/hg
mkdir --parents /tmp/poulpe-clone-cache/hg

echo "POULPE: cloning mercurial from ${repo_url}"

hg clone ${repo_url} poulpe-tmp/hg/mercurial/ \
    --noupdate \
    --config extensions.share= \
    --config share.poolnaming=remote \
    --config share.pool=/tmp/poulpe-clone-cache/hg \
    --config format.use-dirstate-v2=no \
    --config format.use-persistent-nodemap=no \
    --config format.revlog-compression=zlib \
    --config ui.timeout.warn=15

echo "POULPE: clone done"

echo "POULPE: updating to ${MERCURIAL_VERSION}"
# if this looks like a node id pull it to make sure we grab things in topic
if (echo "${MERCURIAL_VERSION}" | egrep -q '^[0-9a-f]{6,64}$'); then
    hg -R poulpe-tmp/hg/mercurial/ pull --rev "${MERCURIAL_VERSION}"
fi
hg -R poulpe-tmp/hg/mercurial/ update "${MERCURIAL_VERSION}"


if [ -n "${evolve_version}" ]; then

    echo "POULPE: cloning evolve from ${evolve_repo_url}"

    hg clone ${evolve_repo_url} poulpe-tmp/hg/evolve/ \
        --noupdate \
        --config extensions.share= \
        --config share.poolnaming=remote \
        --config share.pool=/tmp/poulpe-clone-cache/hg \
        --config format.use-dirstate-v2=no \
        --config format.use-persistent-nodemap=no \
        --config format.revlog-compression=zlib \
        --config ui.timeout.warn=15

    echo "POULPE: clone done"

    echo "POULPE: updating to ${EVOLVE_VERSION}"
    # if this looks like a node id pull it to make sure we grab things in topic
    if (echo "${EVOLVE_VERSION}" | egrep -q '^[0-9a-f]{6,64}$'); then
        hg -R poulpe-tmp/hg/evolve/ pull --rev "${EVOLVE_VERSION}"
    fi
    hg -R poulpe-tmp/hg/evolve/ update "${EVOLVE_VERSION}"
fi

echo "POULPE: building the virtual environment"

$python_override -m venv .

if [ "${flavor}" = "default" ]; then
    EXTRA_INSTALL_ARGS=""
else if [ "${flavor}" = "pure" ]; then
    EXTRA_INSTALL_ARGS="--pure"
else if [ "${flavor}" = "no-rust" ]; then
    EXTRA_INSTALL_ARGS=""
else if [ "${flavor}" = "rust" ]; then
    EXTRA_INSTALL_ARGS="--rust"
else if [ "${flavor}" = "rhg" ]; then
    EXTRA_INSTALL_ARGS="--rust"
else
    echo "POULPE: unknown flavor value: ${flavor}"
fi fi fi fi fi

echo "POULPE: flavor okay ($flavor)"

root="`pwd`"

# temporary work around until we upgrade our packaging
if $python_override --version | grep -q -E "Python 3.1[2-9]" ; then
    "${root}/bin/python" -m pip install "setuptools"
fi

echo "POULPE: installing Mercurial"
(
    cd poulpe-tmp/hg/mercurial/;
    "${root}/bin/python" -m pip install . --quiet --config-settings --global-option=${EXTRA_INSTALL_ARGS} > install-output.txt 2>&1
) || (
    install_ret=$?;
    cat poulpe-tmp/hg/mercurial/install-output.txt >&2;
    exit $install_ret
)

echo "POULPE: install okay"

if [ -n "${evolve_version}" ]; then
    echo "POULPE: installing Evolve"
    (
        cd poulpe-tmp/hg/evolve/;
        "${root}/bin/pip" install . > install-output.txt 2>&1
    ) || (
        install_ret=$?;
        cat poulpe-tmp/hg/evolve/install-output.txt >&2;
        exit $install_ret
    )

    echo "POULPE: install okay"
fi
if [ "$pyre2" == "default" ]; then
    echo "POULPE: no re2 module to install"
else
    "${root}/bin/python" -m pip install "$pyre2"
    echo "POULPE: installed re2 support through: $pyre2"
fi


if [ "${flavor}" == "rhg" ]; then
    echo 'POULPE: building rhg'
    make -C poulpe-tmp/hg/mercurial build-rhg
    mv bin/hg bin/hg.real
    mv poulpe-tmp/hg/mercurial/rust/target/release/rhg bin/hg
    export RHG_ON_UNSUPPORTED="fallback"
    export RHG_FALLBACK_EXECUTABLE="${root}/bin/hg.real"
    echo "POULPE: rhg okay"
fi


echo python.version="`bin/python --version | egrep -o '\S+$'`" > POULPE-VARS
echo python.implementation="`bin/python -c 'import sys; print(sys.implementation.name)'`" >> POULPE-VARS

# XXX we should "simply" take `hg debuginstall -T json` here  (if available, with some filtering)

echo hg.version="`bin/hg --version --quiet| egrep -o '\S+\)$' | egrep -o '[^)]+'`" >> POULPE-VARS
echo hg.flavor="${flavor}" >> POULPE-VARS
echo hg.py-re2-module="${pyre2}" >> POULPE-VARS

for key in node branch tag phase; do
    echo hg.changeset.${key}="`hg log -R poulpe-tmp/hg/mercurial/ -r . -T '{'${key}'}'`" >> POULPE-VARS
done
echo hg.changeset.rank="`hg log -R poulpe-tmp/hg/mercurial/ -r ::. -T 'UNIQUE-TOKEN-FOR-RANK-swilndlkjsagljgalh\n' | grep 'UNIQUE-TOKEN-FOR-RANK-swilndlkjsagljgalh' | wc -l`" >> POULPE-VARS


if [ -e poulpe-tmp/evolve/ ]; then
    echo hg.with-evolve=yes >> POULPE-VARS
    for key in node branch tag phase; do
        echo hg.evolve.changeset.${key}="`hg log -R poulpe-tmp/hg/evolve/ -r . -T '{'${key}'}'`" >> POULPE-VARS
    done
    echo hg.evolve.changeset.rank="`hg log -R poulpe-tmp/hg/evolve/ -r ::. -T 'UNIQUE-TOKEN-FOR-RANK-swilndlkjsagljgalh\n' | grep 'UNIQUE-TOKEN-FOR-RANK-swilndlkjsagljgalh' | wc -l`" >> POULPE-VARS
else
    echo hg.with-evolve=no >> POULPE-VARS
fi

echo "POULPE: cleaning up"

rm -rf poulpe-tmp/hg

echo "POULPE: creating the bin-env.shell file"

cat << EOF > bin-env.shell
#!/bin/bash
set -Eeuo pipefail
this_script=\`realpath \$0\`
export POULPE_BIN_ENV_ROOT="\`dirname \$this_script\`"
. "\${POULPE_BIN_ENV_ROOT}"/bin/activate
export POULPE_SSH_REMOTE_HG="\${POULPE_BIN_ENV_ROOT}/remote-ssh-hg"

# disable local config (this might be the benchmark job instead)
export HGRCPATH="\${POULPE_BIN_ENV_ROOT}/hgrc"
unset HGWITHRUSTEXT
unset HGMODULEPOLICY
EOF

touch hgrc
if [ -n "$evolve_version" ]; then
cat << EOF >> hgrc
[extensions]
evolve=
topic=
EOF
fi

if [ "${flavor}" == "rhg" ]; then
cat << EOF >> bin-env.shell
export RHG_ON_UNSUPPORTED="fallback"
export RHG_FALLBACK_EXECUTABLE="\${POULPE_BIN_ENV_ROOT}/bin/hg.real"
EOF

fi

cat << EOF >> bin-env.shell
"\$@"
EOF

chmod +x bin-env.shell

cat << EOF > remote-ssh-hg
#!/bin/bash
set -Eeuo pipefail
this_script=\`realpath \$0\`
POULPE_BIN_ENV_ROOT="\`dirname \$this_script\`"
\$POULPE_BIN_ENV_ROOT/bin-env.shell hg "\$@"
EOF
chmod +x remote-ssh-hg
