// a more direct translation from the existing shell script
amends "campaign.pkl"

local data: Mapping<String, Listing<DataEnvFilter>(isDistinct)> = new {
    ["modern"] {
        new {key="mercurial.repo.format.compression"; value="zstd"}
    }
    ["all-public"] {
        new {key="mercurial.repo.graph.non-public-visible-revision-count"; value="re:^0$"}
        new {key="mercurial.repo.graph.hidden-revision-count"; value="re:^0$"}
    }
    ["no-rust"] {
        new {key="mercurial.repo.format.persistent-nodemap"; value="no"}
        new {key="mercurial.repo.format.dirstate-v2"; value="no"}
    }
}


tasks {
    for (
        v in List(
            new VersionSpec{mercurial_flavor="no-rust"},
            new VersionSpec{mercurial_flavor="rust"}
        )
    ) {
        //********** test clone for "all public" repository ********************/

        // (default flavor on zstd-sparse-revlog without dirstate-v2 and
        // persistent nodemap)
        //
        // hardlink and stream clone should be relatively
        // fast
        for (protocol in List("local-hardlink", "ssh-stream", "ssh-pull")) {
            new {
                versions = new {v}
                data_env_specs {
                    ...data["modern"]
                    ...data["all-public"]
                    when (!v.is_rust) {
                        ...data["no-rust"]
                    }
                    // pull clone are slower so we skip it for repository over a
                    // million revisions
                    when (protocol.contains("-pull")) {
                        new {
                            key="mercurial.repo.graph.visible-revision-count"
                            value="re:^\\d{0,6}$"
                        }
                    }
                }
                benchmark_name_specs {
                    new {value="re:^hg.command.clone$"}
                }
                benchmark_variant_specs {
                    new {key="protocol"; value=protocol}
                }
            }
        }
        //********** stream/hardlink locked section **********//
        //
        // This should be even faster as doing the stream clone and will watch
        // over and important critical section.
        for (protocol_version in List("v2", "v3")) {
            new {
                versions = new {v}
                data_env_specs {
                    ...data["modern"]
                    ...data["all-public"]
                    when (!v.is_rust) {
                        ...data["no-rust"]
                    }
                }
                benchmark_name_specs {
                    new {value="re:^hg.perf.stream-locked-section$"}
                }
                benchmark_variant_specs {
                    new {key="version"; value=protocol_version}
                }
            }
        }

        //********** pull push benchmarks **********//
        //
        // focus on ssh performance to avoid testing too many things
        //
        // XXX For these, we should also test the non public version too as
        // phase computation and obsmarker impact things.
        for (only_public in List(true, false)) {
            new {
                versions = new {v}
                data_env_specs {
                    ...data["modern"]
                    when(only_public) {
                        ...data["all-public"]
                    }
                    when (!v.is_rust) {
                        ...data["no-rust"]
                    }
                }
                benchmark_name_specs {
                    new {value="re:^hg.command.(push|pull)$"}
                }
                benchmark_variant_specs {
                    new {key="protocol"; value="ssh"}
                    // XXX should we reduce the amount of tested path here ?
                    new {key="explicit"; value="re:.*"}
                    new {key="revs"; value="(none|re:any-(1|100|10000)-extra-rev)"}
                }
            }
        }
        //* Benchmarking `hg log` *//
        for (limit in List(10, 1000)) {
            new {
                versions = new {v}
                data_env_specs {
                    ...data["modern"]
                    ...data["all-public"]
                    when (!v.is_rust) {
                        ...data["no-rust"]
                    }
                }
                benchmark_name_specs {
                    new {value="re:^hg.command.log$"}
                }
                benchmark_variant_specs {
                    new { key="limit-rev"; value="\(limit)"}
                    when ( limit < 1000) {
                        new {key="patch"; value="re:."}
                    }
                }
            }
        }
        //* Benchmarking `hg update` *//
        new {
            versions = new {v}
            data_env_specs {
                ...data["modern"]
                ...data["all-public"]
                when (!v.is_rust) {
                    ...data["no-rust"]
                }
            }
            benchmark_name_specs {
                new {value="re:^hg.command.update$"}
            }
            benchmark_variant_specs {
                new {
                    key="scenario"
                    value="re:^(null-to-tip|1-rev-to-tip|10000-rev-to-tip)$"
                }
            }
        }
    }
    //******************* Benchmark where rhg should be faster **************//
    for (
        v in List(
            new VersionSpec{mercurial_flavor="no-rust"},
            new VersionSpec{mercurial_flavor="rust"},
            new VersionSpec{mercurial_flavor="rhg"}
        )
    ) {
        //* Benchmarking `hg status` *//
        // We test status variant in two dimensions
        for (
            l in List(
                new BenchmarkVariantFilter {key="listed"; value="re:.*"},
                new BenchmarkVariantFilter {key="list-copies"; value="re:.*"}
            )
        ) {
            new {
                versions = new {v}
                data_env_specs {
                    ...data["modern"]
                    ...data["all-public"]
                    when (!v.is_rust) {
                        ...data["no-rust"]
                    }
                }
                benchmark_name_specs {
                    new {value="re:^hg.command.status$"}
                }
                // XXX we need to test more state variant
                benchmark_variant_specs {
                    new {key="repo-state"; value="clean"}
                    l
                }
            }
        }
        //* Benchmarking `hg status` between revision *//
            new {
                versions = new {v}
                data_env_specs {
                    ...data["modern"]
                    ...data["all-public"]
                    when (!v.is_rust) {
                        ...data["no-rust"]
                    }
                }
                benchmark_name_specs {
                    new {value="re:^hg.command.status$"}
                }
                benchmark_variant_specs {
                    new {key="list-copies"; value="re:.*"}
                    // the repository state is not really relevant here
                    new {key="repo-state"; value="clean"}
                    // XXX consider reducing the number of case tested.
                    // the from-rev-0-to-tip is important for copy tracing
                    new {key="rage"; value="re:^from-rev-0-to-tip|from-1000th-tiprev-to-tip|change-tip|change-1000th-tiprev$"}
                }
            }

        //* Benchmarking `hg cat` *//
        new {
            versions = new {v}
            data_env_specs {
                ...data["modern"]
                ...data["all-public"]
                when (!v.is_rust) {
                    ...data["no-rust"]
                }
            }
            benchmark_name_specs {
                new {value="re:^hg.command.cat$"}
            }
            benchmark_variant_specs {
                new {key="files"; value="re:^all-root$"}
            }
        }
        //* Benchmarking `hg files` *//
        new {
            versions = new {v}
            data_env_specs {
                ...data["modern"]
                ...data["all-public"]
                when (!v.is_rust) {
                    ...data["no-rust"]
                }
            }
            benchmark_name_specs {
                new {value="re:^hg.command.files$"}
            }
            benchmark_variant_specs {
                new {key="files"; value="re:^default|all-list-filepath-shuffled$"}
            }
        }
        //* Benchmarking annotate *//
        new {
            versions = new {v}
            data_env_specs {
                ...data["modern"]
                ...data["all-public"]
                when (!v.is_rust) {
                    ...data["no-rust"]
                }
            }
            benchmark_name_specs {
                new {value="re:^hg.command.annotate$"}
            }
            benchmark_variant_specs {
                new {key="files"; value="re:.*"}
                new {key="follow-copies"; value="re:^yes|no$"}
            }
        }
    }
}
