#!/usr/bin/env python3
"""Script to setup a data environment for working-copy-related benchmarking
"""
import copy
import os
import pathlib
import shutil
import subprocess
import tempfile
from typing import List, Union

import click

import poulpe
from poulpe.dataenvs import DataEnvironment


def get_minimum_hg_env():
    """Return the smallest viable environment for `hg` suprocesses"""
    env_path = os.environ.get("PATH")
    env = {"HGPLAIN": "1", "HGRCPATH": "1"}
    if env_path:
        # Sometimes (in tests for example), there is no PATH.
        # An empty PATH could be interpreted differently than a lack of
        # PATH by some programs.
        env["PATH"] = env_path
    return env


def call_hg(cmd: List[Union[str, bytes]], hg="hg", env_override=None):
    env = get_minimum_hg_env()
    if env_override is not None:
        env.update(env_override)

    return subprocess.check_call(
        [hg] + cmd,
        env=env,
    )


def brick_apply_hg_copies(repo_path, brick_file, brick_temp_dir, params):
    data = (brick_temp_dir / params["files_pairs_list"]).read_bytes()
    list_of_pairs = data.split(b"\n")
    minimal_env = get_minimum_hg_env()

    for pair in list_of_pairs:
        if not pair:
            continue
        source, target = pair.split(b"\0")

        call_hg(["cp", "--cwd", repo_path, source, target])


def brick_apply_hg_remove(repo_path, brick_file, brick_temp_dir, params):
    data = (brick_temp_dir / params["files_list"]).read_bytes()
    list_of_files = data.split(b"\n")

    for file in list_of_files:
        if not file:
            continue

        call_hg(["rm", "--cwd", repo_path, file])


def brick_apply_hg_add(repo_path, brick_file, brick_temp_dir, params):
    data = (brick_temp_dir / params["files_list"]).read_bytes()
    list_of_files = data.split(b"\n")

    for file in list_of_files:
        if not file:
            continue

        call_hg(["add", "--cwd", repo_path, file])


appliers = {
    **poulpe.dataenvs.brick_appliers,
    "hg::working-copy.add-files": brick_apply_hg_add,
    "hg::working-copy.copy-files": brick_apply_hg_copies,
    "hg::working-copy.remove-files": brick_apply_hg_remove,
}


def repo_name_for_brick_item(main, item):
    main = bytes(main).replace(b"-main", b"")
    return b"%s-%s" % (main, os.fsencode(item))


def copy_repo(main, target_name, revision):
    new_repo = repo_name_for_brick_item(main, target_name)
    click.echo(f"Copying repo {main} to {new_repo}")

    subprocess.check_call(["cp", "-a", "--reflink=auto", main, new_repo])

    update_repo_to_revision(new_repo, revision)
    return pathlib.Path(os.fsdecode(new_repo))


def update_repo_to_revision(repo, revision):
    click.echo(f"Updating to revision {revision}")
    return call_hg(["-R", repo, "up", revision, "--quiet"])


def get_list_matcher():
    # Matches an expected "<name>.list" to actions to take with those lists
    base_matcher = {
        "ignored": [["poulpe::create-files", {"files_list": "ignored.list"}]],
        "unknowns": [["poulpe::create-files", {"files_list": "unknowns.list"}]],
        "added": [
            ["poulpe::create-files", {"files_list": "added.list"}],
            ["hg::working-copy.add-files", {"files_list": "added.list"}],
        ],
        "modified": [
            [
                "poulpe::modify-files",
                {"files_list": "modified.list"},
            ],
        ],
        "removed": [
            [
                "hg::working-copy.remove-files",
                {"files_list": "removed.list"},
            ],
        ],
        "copied": [
            [
                "hg::working-copy.copy-files",
                {"files_pairs_list": "copied.list"},
            ],
        ],
        "deleted": [
            ["poulpe::delete-files", {"files_list": "deleted.list"}],
        ],
    }
    full_matcher = {}
    for name, actions in base_matcher.items():
        full_matcher[name] = actions
        actions_for_large = []
        for applier_name, params in actions:
            new_params = copy.deepcopy(params)
            if params.get("files_list") is not None:
                new_params["files_list"] = f'large-{params["files_list"]}'
            if params.get("files_pairs_list") is not None:
                new_params[
                    "files_pairs_list"
                ] = f'large-{params["files_pairs_list"]}'
            actions_for_large.append([applier_name, new_params])

        full_matcher[f"large-{name}"] = actions_for_large

    return full_matcher


# Keep in sync with working-copy related benchmarks
NON_DEFAULT_BENCHMARK_VARS = {
    "large-ignored": {"large-ignored"},
    "large-all": {
        "large-ignored",
        "large-unknowns",
        "large-added",
        "large-modified",
        "large-deleted",
        "large-removed",
        "large-copied",
    },
    "all-and-large-ignored": {
        "large-ignored",
        "unknowns",
        "added",
        "modified",
        "deleted",
        "removed",
        "copied",
    },
    "modified-and-copied": {
        "copied",
        "modified",
    },
    "modified": {
        "modified",
    },
    "large-modified-and-copied": {
        "large-copied",
        "large-modified",
    },
    "large-modified": {
        "large-modified",
    },
    "ignored": {"ignored"},
    "ignored-unknowns": {"ignored", "unknowns"},
}


def apply_brick(data_env: DataEnvironment, brick_file: pathlib.Path):
    click.echo(f"Unpacking brick {brick_file}")
    # TODO generalize and move to poulpe core
    # TODO figure out hg vs mercurial in the whole of Poulpe
    tmpdir = tempfile.mkdtemp()
    try:
        if brick_file.is_dir():
            # Maybe it's already unpacked
            brick_dir = brick_file
        else:
            brick_dir = pathlib.Path()
            shutil.unpack_archive(brick_file, brick_dir)

        apply_brick_inner(data_env, brick_file, brick_dir)
    finally:
        shutil.rmtree(tmpdir, ignore_errors=True)


def apply_brick_inner(data_env, brick_file, brick_dir):
    brick_info = poulpe.basics.get_data(brick_dir / "data-brick.poulpe")
    if brick_info is None:
        raise poulpe.errors.InvalidBrick(
            "missing data-brick.poulpe description file", brick_file
        )

    brick_kind = poulpe.basics.get_one_value(brick_info, "brick.kind")
    if brick_kind is None:
        raise poulpe.errors.InvalidBrick("missing brick.kind key", brick_file)
    if brick_kind != "hg::working-copy":
        raise poulpe.errors.InvalidBrick(
            f"expected kind of 'hg::working-copy', got '{brick_kind}'",
            brick_file,
        )

    brick_name = poulpe.basics.get_one_value(brick_info, "brick.name")
    if brick_name is None:
        raise poulpe.errors.InvalidBrick("missing brick.name key", brick_file)

    key = "hg.update-to-revision"
    revision = poulpe.basics.get_one_value(brick_info, key)
    if revision is None:
        raise poulpe.errors.InvalidBrick(
            f"missing {key} key. (working copy bricks only apply to one rev)",
            brick_file,
        )

    items = poulpe.basics.get_one_value(brick_info, "items")
    if not items:
        raise poulpe.errors.InvalidBrick("empty items", brick_file)
    items = set(items)

    applied_brick_items = []

    main_repo_path = data_env.get_benchmark_input_vars(
        "mercurial.main-repo-path"
    )

    update_repo_to_revision(
        pathlib.Path(data_env.path, main_repo_path), revision
    )
    data_env.set_benchmark_input_var(
        "mercurial.working-copy.update-to-revision", revision
    )

    list_matcher = get_list_matcher()

    for benchmark_var, required_lists in sorted(
        NON_DEFAULT_BENCHMARK_VARS.items()
    ):
        missing = required_lists - items
        data_var = f"mercurial.working-copy.{benchmark_var}"
        if len(missing) == 0:
            try:
                repo_path = copy_repo(
                    pathlib.Path(data_env.path, main_repo_path),
                    data_var,
                    revision,
                )
                for l in sorted(required_lists):
                    for applier_name, params in list_matcher[l]:
                        applier = appliers[applier_name]
                        click.echo(
                            f"Applying {applier_name} with params {params}"
                        )
                        applier(repo_path, brick_file, brick_dir, params)
                applied_brick_items.append(data_var)
            except:
                if repo_path is not None:
                    shutil.rmtree(repo_path, ignore_errors=True)
                raise
        else:
            missing = sorted(missing)
            click.secho(
                f"'{benchmark_var}' requires missing lists: '{missing}', skipping.",
                fg="yellow",
            )

    if not applied_brick_items:
        raise poulpe.errors.InvalidBrick(
            "Not enough lists to build any repo, aborting."
        )

    data_env.set_applied_brick_items(brick_name, applied_brick_items)
    for item in applied_brick_items:
        # This jumping around encoding may not be worth it
        # or at least should be cleaned up
        data_env.set_benchmark_input_var(
            item,
            os.fsdecode(
                repo_name_for_brick_item(os.fsencode(main_repo_path), item)
            ),
        )


@click.command()
@click.argument("data-env-path")
@click.argument("data-brick-path", required=False)
def setup_working_copy(data_env_path, data_brick_path):
    data_env = poulpe.get_data_env(data_env_path)
    main_repo_path = data_env.get_benchmark_input_vars(
        "mercurial.main-repo-path"
    )
    data_env.set_benchmark_input_var(
        "mercurial.working-copy.clean", main_repo_path
    )

    if data_brick_path is not None:
        brick_file = pathlib.Path(data_brick_path)
        apply_brick(data_env, brick_file)
    else:
        click.secho(
            f"Warning: running setup-working-copy without a data brick",
            fg="yellow",
        )
        click.secho(
            f"(non-default working-copy variants will be unavailable)",
            fg="yellow",
        )

    data_env.write()


if __name__ == "__main__":
    setup_working_copy()
