#!/usr/bin/env python3
"""Find suitable revision to run delta-find benchmark on
"""
import json
import os
import pathlib
import subprocess
import sys

import poulpe.dataenvs

if len(sys.argv) != 2:
    script_name = os.path.basename(sys.argv[0])
    print(f"USAGE: {script_name} DATA-ENV-PATH", file=sys.stderr)
    print("  DATA-ENV-PATH: the path to that data-env", file=sys.stderr)
    sys.exit(32)

data_env_directory = pathlib.Path(sys.argv[1])
desc_path = data_env_directory / "data-env.poulpe"

data_env = poulpe.dataenvs.get_data_env(data_env_directory)

main_repo = data_env.get_benchmark_input_vars("mercurial.main-repo-path")

repo_path = data_env_directory / main_repo

cmd = [
    "hg",
    "--repository",
    repo_path,
    "debugdeltachain",
    "--template",
    "json",
    "--manifest",
]
proc = subprocess.run(cmd, stdout=subprocess.PIPE, encoding="utf8")
proc.check_returncode()
delta_chain_data = json.loads(proc.stdout)


all_info = {}
all_snap_rev = []
some_snap = {}
all_merge = []

target_type = ("base", "snap")
for rev_info in delta_chain_data:
    all_info[rev_info["rev"]] = rev_info
    if rev_info["p1"] != -1 and rev_info["p2"] != -1:
        all_merge.append(rev_info)
    if rev_info["deltatype"] not in target_type:
        continue
    all_snap_rev.append(rev_info["rev"])


for rev in all_snap_rev[-100:]:
    cmd = [
        "hg",
        "--repository",
        repo_path,
        "debug-delta-find",
        "--manifest",
        "%d" % rev,
    ]
    proc = subprocess.run(cmd, stdout=subprocess.PIPE, encoding="utf8")
    proc.check_returncode()
    data = None
    for line in proc.stdout.splitlines():
        if line.startswith("DBG-DELTAS: MANIFESTLOG:"):
            pieces = line.split()
            data = {}
            for p in pieces:
                if "=" in p:
                    k, v = p.split("=", 1)
                    if "." in v:
                        try:
                            v = float(v)
                        except ValueError:
                            pass
                    else:
                        try:
                            v = int(v)
                        except ValueError:
                            pass
                    data[k] = v
            break
    if data is not None:
        some_snap[rev] = data

items = list(some_snap.items())


def set_one(key, value):
    cmd = [
        "poulpe",
        "env-desc",
        "set",
        str(desc_path),
        f"bench-input-vars.mercurial.revlog.delta-find.manifest.{key}",
        str(value),
    ]
    subprocess.check_call(cmd)


# search snapshot with the most total try
items.sort(key=lambda i: (i[1]["try-count"], i[0]))
max_tries = items[:-6:-1]
for i, r in enumerate(max_tries):
    c = chr(ord("a") + i)
    set_one(f"snapshot.many-tries.{c}", r[0])

# search snapshot with the most search-rounds
items.sort(key=lambda i: (i[1]["search-rounds"], i[0]))
max_round = items[:-6:-1]
for i, r in enumerate(max_round):
    c = chr(ord("a") + i)
    set_one(f"snapshot.many-rounds.{c}", r[0])

# search snapshot with the most `try * rounds`
items.sort(key=lambda i: (i[1]["try-count"] * i[1]["search-rounds"], i[0]))
max_combined = items[:-6:-1]
for i, r in enumerate(max_combined):
    c = chr(ord("a") + i)
    set_one(f"snapshot.many-both.{c}", r[0])


# search merges listing on the longest chain for parent
def key(i):
    p1_len = all_info[i["p1"]]["chainlen"]
    p2_len = all_info[i["p2"]]["chainlen"]
    return (min(p1_len, p2_len), max(p1_len, p2_len), i["rev"])


all_merge.sort(key=key)
max_merge = all_merge[:-6:-1]
for i, r in enumerate(max_merge):
    c = chr(ord("a") + i)
    set_one(f"merge.long-chain-both.{c}", r["rev"])
