#!/usr/bin/env python3
from pathlib import Path
import subprocess
import sys
from typing import Optional
import click

SCRIPT_DIR = Path(__file__).parent


@click.command(no_args_is_help=True)
@click.argument(
    "data_env",
    type=click.Path(
        exists=True,
        dir_okay=True,
        file_okay=False,
        resolve_path=True,
        path_type=Path,
    ),
)
@click.option(
    "--repo",
    type=click.Path(path_type=Path),
    default="main",
    help="Path of the main repository inside the data-env (defaults to 'main')",
)
@click.option(
    "--name",
    help="Name of the data-environment in results (defaults to the folder name)",
)
@click.option(
    "--wc-brick",
    type=click.Path(exists=True, dir_okay=True, file_okay=True, path_type=Path),
    help="Path to a working-copy data brick for this data-env\n(See 'suites/hg/make-working-copy-brick.py')",
)
@click.option(
    "--bundle-brick",
    type=click.Path(exists=True, dir_okay=True, file_okay=True, path_type=Path),
    help="Path to a bundles data brick for this data-env\n(See 'suites/hg/brick-setup/create-bundle-brick')",
)
@click.option(
    "--force",
    is_flag=True,
    help="Overwrites the data-env description file if it exists",
)
def setup_full(
    data_env: Path,
    repo: Path,
    name,
    wc_brick,
    bundle_brick,
    force: bool,
):
    """Record information about a data environment, given a folder
    and a reference repository.

    This reference repo may be used to derive other repos inside the same
    data-env.
    For example, most working-copy benchmarks need a repo with an
    "unclean" working copy, which implies controlling precisely
    the repo state (like mtimes, etc.) when building the variant.

    \b
    ARGS:
        DATA_ENV: Path to the data-env folder containing the main repository
    """
    if isinstance(data_env, bytes):
        data_env = data_env.decode("utf8")
        data_env = Path(data_env)
    description_file = data_env / "data-env.poulpe"
    if description_file.exists():
        if force:
            description_file.unlink(missing_ok=True)
        else:
            click.secho(
                "data-env already has a description file, aborting",
                err=True,
                fg="red",
            )
            click.secho(
                "hint: use --force to process anyway and overwrite",
                err=True,
                fg="yellow",
            )
            sys.exit(1)

    if name is None:
        name = data_env.name

    if isinstance(repo, bytes):
        repo = repo.decode("utf8")
    relative_repo = data_env / repo
    if not relative_repo.is_dir():
        click.secho(
            "invalid repo argument, not a directory inside the data-env",
            err=True,
            fg="red",
        )
        sys.exit(1)

    data_env = str(data_env)
    repo = str(repo)
    wc_brick = str(wc_brick) if wc_brick is not None else None
    bundle_brick = str(bundle_brick) if bundle_brick is not None else None

    # XXX should probably automatically find scripts and run them.
    # However there are few of them now so explicit listing seems simpler.

    click.echo("Creating the base data-env")
    subprocess.run(
        [SCRIPT_DIR / "setup-00-minimal.sh", name, data_env, repo], check=True
    )

    click.echo("Registering graph details")
    subprocess.run([SCRIPT_DIR / "setup-01-graph.sh", data_env], check=True)

    click.echo("Registering the format details")
    subprocess.run([SCRIPT_DIR / "setup-10-format.sh", data_env], check=True)

    click.echo("Setting up working-copy benchmarks")
    if wc_brick is not None:
        subprocess.run(
            [SCRIPT_DIR / "setup-50-working-copy.py", data_env, wc_brick],
            check=True,
        )
    else:
        subprocess.run(
            [SCRIPT_DIR / "setup-50-working-copy.py", data_env],
            check=True,
        )

    click.echo("Setting up `delta-find` benchmark (can take a very long time)")
    subprocess.run(
        [SCRIPT_DIR / "setup-55-delta-chain.py", data_env], check=True
    )
    if bundle_brick is not None:
        click.echo("Setting up bundles related benchmark")
        subprocess.run(
            [SCRIPT_DIR / "setup-60-bundles", data_env, bundle_brick],
            check=True,
        )

    click.echo("Setting up `annotate` benchmark (can take up to a few minutes)")
    subprocess.run(
        [SCRIPT_DIR / "setup-70-annotate.sh", data_env],
        check=True,
    )


if __name__ == "__main__":
    setup_full()
